import numpy as np
from skimage import segmentation
from scipy.ndimage import distance_transform_edt
import os.path as osp
import glob 
import os
import cv2



def ex_find_boundries(labels):
    boundaries = segmentation.find_boundaries(labels, mode='thick')
    print("thick: \n", boundaries.astype(np.uint8))
    '''
        [[0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 1 1 1 0 0]
        [0 0 1 1 1 1 1 1 1 0]
        [0 1 1 1 1 1 0 1 1 0]
        [0 1 1 0 1 1 0 1 1 0]
        [0 1 1 1 1 1 0 1 1 0]
        [0 0 1 1 1 1 1 1 1 0]
        [0 0 0 0 0 1 1 1 0 0]
        [0 0 0 0 0 0 0 0 0 0]]
    '''
    
    boundaries = segmentation.find_boundaries(labels, mode='inner')
    print("inner: \n", boundaries.astype(np.uint8))
    '''
        [[0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 1 1 1 0 0]
        [0 0 1 1 1 1 0 1 0 0]
        [0 0 1 0 1 1 0 1 0 0]
        [0 0 1 1 1 1 0 1 0 0]
        [0 0 0 0 0 1 1 1 0 0]
        [0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0]]
    '''

    boundaries = segmentation.find_boundaries(labels, mode='outer')
    print("outer: \n", boundaries.astype(np.uint8))
    '''
        [[0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 1 1 1 0 0]
        [0 0 1 1 1 1 0 0 1 0]
        [0 1 0 0 1 1 0 0 1 0]
        [0 1 0 0 1 1 0 0 1 0]
        [0 1 0 0 1 1 0 0 1 0]
        [0 0 1 1 1 1 0 0 1 0]
        [0 0 0 0 0 1 1 1 0 0]
        [0 0 0 0 0 0 0 0 0 0]]
    '''
    
def ex_distance_transform_edt(labels, tolerance):
    
    boundaries = segmentation.find_boundaries(labels, mode='thick')
    print("~boundaries: \n", (~boundaries).astype(np.uint8))
    '''
        [[1 1 1 1 1 1 1 1 1 1]
        [1 1 1 1 1 0 0 0 1 1]
        [1 1 0 0 0 0 0 0 0 1]
        [1 0 0 0 0 0 1 0 0 1]
        [1 0 0 1 0 0 1 0 0 1]
        [1 0 0 0 0 0 1 0 0 1]
        [1 1 0 0 0 0 0 0 0 1]
        [1 1 1 1 1 0 0 0 1 1]
        [1 1 1 1 1 1 1 1 1 1]]
    '''
    edt = distance_transform_edt(~boundaries)
    print("\nedt: \n", edt.astype(np.uint8))
    '''
        [[2 2 2 2 1 1 1 1 1 2]
        [2 1 1 1 1 0 0 0 1 1]
        [1 1 0 0 0 0 0 0 0 1]
        [1 0 0 0 0 0 1 0 0 1]
        [1 0 0 1 0 0 1 0 0 1]
        [1 0 0 0 0 0 1 0 0 1]
        [1 1 0 0 0 0 0 0 0 1]
        [2 1 1 1 1 0 0 0 1 1]
        [2 2 2 2 1 1 1 1 1 2]]
    '''
    
    edt = edt <= tolerance
    print("\nedt <= tolerance: \n", edt.astype(np.uint8))
    '''
        [[0 0 0 0 0 1 1 1 0 0]
        [0 0 1 1 1 1 1 1 1 0]
        [0 1 1 1 1 1 1 1 1 1]
        [1 1 1 1 1 1 1 1 1 1]
        [1 1 1 1 1 1 1 1 1 1]
        [1 1 1 1 1 1 1 1 1 1]
        [0 1 1 1 1 1 1 1 1 1]
        [0 0 1 1 1 1 1 1 1 0]
        [0 0 0 0 0 1 1 1 0 0]]
    '''
  
def bfscore(pred, gt, tolerances=None, mode="inner"):
    """
    - True Positive (TP): 예측 경계와 실제 경계가 허용 오차 내에 있을 때.
    - False Positive (FP): 예측 경계가 실제 경계 근처에 없을 때.
    - False Negative (FN): 실제 경계가 예측 경계 근처에 없을 때.
    """

    if isinstance(tolerances, int):
        tolerances = {"pred": tolerances, "gt": tolerances}
    elif isinstance(tolerances, dict):
        assert "pred" in tolerances, ValueError(f"Tolerance must habe pred")
        assert "gt" in tolerances, ValueError(f"Tolerance must habe gt")
    else:
        tolerances = {"pred": 1, "gt": 1}

    pred_boundaries = segmentation.find_boundaries(pred, mode=mode)
    gt_boundaries = segmentation.find_boundaries(gt, mode=mode)

    pred_edt = distance_transform_edt(~pred_boundaries)
    gt_edt = distance_transform_edt(~gt_boundaries)

    pred_edt = pred_edt <= tolerances["pred"]
    gt_edt = gt_edt <= tolerances["gt"]

    tp = np.sum(pred_boundaries & gt_edt)
    fp = np.sum(pred_boundaries & ~gt_edt)
    fn = np.sum(gt_boundaries & ~pred_edt)

    precision = tp / (tp + fp + 1e-7)
    recall = tp / (tp + fn + 1e-7)
    f1_score = 2 * (precision * recall) / (precision + recall + 1e-7)

    return {"f1_score": f1_score, "precision": precision, "recall": recall}


def bfscores_by_channel(pred, gt, tolerances=None, mode="inner"):
    assert pred.ndim == 3, ValueError(f"Pred must be 3D, not {pred.shape}")
    assert gt.ndim == 3, ValueError(f"Gt must be 3D, not {gt.shape}")
    assert pred.shape[-1] == gt.shape[-1], ValueError(
        f"Pred({pred.shape[-1]}) and Gt({gt.shape[-1]}) must have same number of channels"
    )

    scores = {}
    for ch_idx in range(1, pred.shape[-1]):
        score = bfscore(
            pred[:, :, ch_idx], gt[:, :, ch_idx], tolerances=tolerances, mode=mode
        )
        scores[ch_idx] = {
            "f1_score": score["f1_score"],
            "precision": score["precision"],
            "recall": score["recall"],
        }

    return scores


def bfscores_by_batch(pred, gt, tolerances=None, mode="inner"):

    pred = pred.numpy()
    gt = gt.numpy()

    assert pred.ndim == 4, ValueError(f"Pred must be 4D, not {pred.shape}")
    assert gt.ndim == 4, ValueError(f"Gt must be 4D, not {gt.shape}")
    assert pred.shape[-1] == gt.shape[-1], ValueError(
        f"Pred({pred.shape[-1]}) and Gt({gt.shape[-1]}) must have same number of channels"
    )

    total_scores = []
    for batch_idx in range(0, pred.shape[0]):
        scores = bfscores_by_channel(
            pred[batch_idx], gt[batch_idx], tolerances=tolerances, mode=mode
        )
        total_scores.append(scores)

    scores = {}
    for total_score in total_scores:
        for label, score in total_score.items():
            scores[label] = {
                "f1_score": tf.convert_to_tensor(np.mean(score["f1_score"])),
                "precision": tf.convert_to_tensor(np.mean(score["precision"])),
                "recall": tf.convert_to_tensor(np.mean(score["recall"])),
            }

    return scores


    
if __name__ == '__main__':
    # ## For example ---------------------------------------------------------- 
    # labels = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    #                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    #                     [0, 0, 0, 0, 0, 5, 5, 5, 0, 0],
    #                     [0, 0, 1, 1, 1, 5, 5, 5, 0, 0],
    #                     [0, 0, 1, 1, 1, 5, 5, 5, 0, 0],
    #                     [0, 0, 1, 1, 1, 5, 5, 5, 0, 0],
    #                     [0, 0, 0, 0, 0, 5, 5, 5, 0, 0],
    #                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    #                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]], dtype=np.uint8)

    # ex_find_boundries(labels)
    # ex_distance_transform_edt(labels, tolerance=0)
    # ## ----------------------------------------------------------------------

    img_file = '/HDD/_projects/etc/bfscore_python-master/data/pred_0.png'
    mask_file = '/HDD/_projects/etc/bfscore_python-master/data/gt_0.png'
    assert osp.exists(img_file), ValueError(f"There is no such image: {img_file}")
    assert osp.exists(mask_file), ValueError(f"There is no such image: {mask_file}")

    tolerance = 2

    pred = cv2.imread(img_file)
    gt = cv2.imread(mask_file)

    scores = bfscores_by_channel(pred, gt)
    print(scores)


from visionsuite.engines.segmentation.src.loops.build import build_loop
from visionsuite.engines.segmentation.src.datasets.build import build_dataset
from visionsuite.engines.segmentation.src.models.build import build_model
from visionsuite.engines.segmentation.utils.registry import RUNNERS
from visionsuite.engines.utils.bases import BaseTestRunner
from visionsuite.engines.utils.callbacks import Callbacks
from .callbacks import test_callbacks

@RUNNERS.register()
class TestRunner(BaseTestRunner, Callbacks):
    def __init__(self, task):
        BaseTestRunner.__init__(self, name='TestRunner', task=task)
        Callbacks.__init__(self)
        
        self.add_callbacks(test_callbacks)

    def set_configs(self, *args, **kwargs):
        super().set_configs(mode='test', *args, **kwargs)
        
        self.run_callbacks('on_runner_set_configs')
        
    def set_variables(self):
        super().set_variables()
        
        self.run_callbacks('on_runner_set_variables')
    
    def run(self):
        super().run()
        
        from visionsuite.engines.segmentation.src.datasets.mask_dataset import get_transform
        transform = get_transform(False, {"weights": None, "test_only": False, "backend": 'PIL', "use_v2": False})
        self.run_callbacks('on_runner_run_start')
        dataset = build_dataset(**self.args['test']['dataset'], transform=transform)
        self.log_info(f"Dataset is LOADED and BUILT", self.run.__name__, __class__.__name__)
        
        model = build_model(**self.args['model'])
        model.build(**self.args['model'], 
                    num_classes=len(self.args['test']['dataset']['classes']), 
                    train=self.args['test'], 
                    distributed=self.args['distributed']['use'],
                    _logger=self.args['model'].get('logger', None),
                    seed_model=self.args['test']['seed_model']
        )
        self.log_info(f"Model is LOADED and BUILT", self.run.__name__, __class__.__name__)
        
        loop = build_loop(**self.args['test']['loop'])
        loop.build(_model=model, 
                   _dataset=dataset, 
                   _archive=self._archive, 
                   **self.args,
                   _logger=self.args['loop'].get('logger', None)
                )
        self.log_info(f"Loop is LOADED and BUILT", self.run.__name__, __class__.__name__)

        loop.run()
        
        self.run_callbacks('on_runner_run_end')



import torch

from visionsuite.engines.utils.torch_utils.dist import reduce_across_processes

class ConfusionMatrix:
    def __init__(self, num_classes):
        self.num_classes = num_classes
        self.mat = None
        self.values = {}

    def update(self, a, b):
        n = self.num_classes
        if self.mat is None:
            self.mat = torch.zeros((n, n), dtype=torch.int64, device=a.device)
        with torch.inference_mode():
            k = (a >= 0) & (a < n)
            inds = n * a[k].to(torch.int64) + b[k]
            self.mat += torch.bincount(inds, minlength=n**2).reshape(n, n)

    def reset(self):
        self.mat.zero_()

    def compute(self):
        h = self.mat.float()
        acc_global = torch.diag(h).sum() / h.sum()
        acc = torch.diag(h) / h.sum(1)
        iu = torch.diag(h) / (h.sum(1) + h.sum(0) - torch.diag(h))
        
        self.update_values(acc_global, acc, iu)
        
        return acc_global, acc, iu

    def update_values(self, acc_global, acc, iu):
        self.values.update({"acc_global": acc_global.item()*100})
        self.values.update({ f"{idx}_acc": _acc for idx, _acc in enumerate((acc * 100).tolist())})
        self.values.update({ f"{idx}_iou": _iou for idx, _iou in enumerate((iu * 100).tolist())})
        self.values.update({"mean IoU": iu.mean().item() * 100})


    def reduce_from_all_processes(self):
        self.mat = reduce_across_processes(self.mat).to(torch.int64)

    def __str__(self):
        acc_global, acc, iu = self.compute()
        return ("global correct: {:.1f}\naverage row correct: {}\nIoU: {}\nmean IoU: {:.1f}").format(
            acc_global.item() * 100,
            [f"{i:.1f}" for i in (acc * 100).tolist()],
            [f"{i:.1f}" for i in (iu * 100).tolist()],
            iu.mean().item() * 100,
        )


import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
from scipy.ndimage import distance_transform_edt as distance
# can find here: https://github.com/CoinCheung/pytorch-loss/blob/af876e43218694dc8599cc4711d9a5c5e043b1b2/label_smooth.py
from .label_smooth import LabelSmoothSoftmaxCEV1 as LSSCE
from torchvision import transforms
from functools import partial
from operator import itemgetter

# Tools
def kl_div(a,b): # q,p
    return F.softmax(b, dim=1) * (F.log_softmax(b, dim=1) - F.log_softmax(a, dim=1))   

def one_hot2dist(seg):
    res = np.zeros_like(seg)
    for i in range(len(seg)):
        posmask = seg[i].astype(bool)
        if posmask.any():
            negmask = ~posmask
            res[i] = distance(negmask) * negmask - (distance(posmask) - 1) * posmask
    return res

def class2one_hot(seg, C):
    seg = seg.unsqueeze(dim=0) if len(seg.shape) == 2 else seg
    res = torch.stack([seg == c for c in range(C)], dim=1).type(torch.int32)
    return res

# Active Boundary Loss
class ABL(nn.Module):
    def __init__(self, isdetach=True, max_N_ratio = 1/100, ignore_label = 255, label_smoothing=0.2, weight = None, max_clip_dist = 20.):
        super(ABL, self).__init__()
        self.ignore_label = ignore_label
        self.label_smoothing = label_smoothing
        self.isdetach=isdetach
        self.max_N_ratio = max_N_ratio

        self.weight_func = lambda w, max_distance=max_clip_dist: torch.clamp(w, max=max_distance) / max_distance

        self.dist_map_transform = transforms.Compose([
            lambda img: img.unsqueeze(0),
            lambda nd: nd.type(torch.int64),
            partial(class2one_hot, C=1),
            itemgetter(0),
            lambda t: t.cpu().numpy(),
            one_hot2dist,
            lambda nd: torch.tensor(nd, dtype=torch.float32)
        ])

        if label_smoothing == 0:
            self.criterion = nn.CrossEntropyLoss(
                weight=weight,
                ignore_index=ignore_label,
                reduction='none'
            )
        else:
            self.criterion = LSSCE(
                reduction='none',
                ignore_index=ignore_label,
                lb_smooth = label_smoothing
            )

    def logits2boundary(self, logit):
        eps = 1e-5
        _, _, h, w = logit.shape
        max_N = (h*w) * self.max_N_ratio
        kl_ud = kl_div(logit[:, :, 1:, :], logit[:, :, :-1, :]).sum(1, keepdim=True)
        kl_lr = kl_div(logit[:, :, :, 1:], logit[:, :, :, :-1]).sum(1, keepdim=True)
        kl_ud = torch.nn.functional.pad(
            kl_ud, [0, 0, 0, 1, 0, 0, 0, 0], mode='constant', value=0)
        kl_lr = torch.nn.functional.pad(
            kl_lr, [0, 1, 0, 0, 0, 0, 0, 0], mode='constant', value=0)
        kl_combine = kl_lr+kl_ud
        while True: # avoid the case that full image is the same color
            kl_combine_bin = (kl_combine > eps).to(torch.float)
            if kl_combine_bin.sum() > max_N:
                eps *=1.2
            else:
                break
        #dilate
        # dilate_weight = torch.ones((1,1,3,3)).cuda()
        dilate_weight = torch.ones((1,1,3,3))
        edge2 = torch.nn.functional.conv2d(kl_combine_bin, dilate_weight, stride=1, padding=1)
        edge2 = edge2.squeeze(1)  # NCHW->NHW
        kl_combine_bin = (edge2 > 0)
        return kl_combine_bin

    def gt2boundary(self, gt, ignore_label=-1):  # gt NHW
        gt_ud = gt[:,1:,:]-gt[:,:-1,:]  # NHW
        gt_lr = gt[:,:,1:]-gt[:,:,:-1]
        gt_ud = torch.nn.functional.pad(gt_ud, [0,0,0,1,0,0], mode='constant', value=0) != 0 
        gt_lr = torch.nn.functional.pad(gt_lr, [0,1,0,0,0,0], mode='constant', value=0) != 0
        gt_combine = gt_lr+gt_ud
        del gt_lr
        del gt_ud
        
        # set 'ignore area' to all boundary
        gt_combine += (gt==ignore_label)
        
        return gt_combine > 0

    def get_direction_gt_predkl(self, pred_dist_map, pred_bound, logits):
        # NHW,NHW,NCHW
        eps = 1e-5
        # bound = torch.where(pred_bound)  # 3k
        bound = torch.nonzero(pred_bound*1)
        n,x,y = bound.T
        max_dis = 1e5

        logits = logits.permute(0,2,3,1) # NHWC

        pred_dist_map_d = torch.nn.functional.pad(pred_dist_map,(1,1,1,1,0,0),mode='constant', value=max_dis) # NH+2W+2

        logits_d = torch.nn.functional.pad(logits,(0,0,1,1,1,1,0,0),mode='constant') # N(H+2)(W+2)C
        logits_d[:,0,:,:] = logits_d[:,1,:,:] # N(H+2)(W+2)C
        logits_d[:,-1,:,:] = logits_d[:,-2,:,:] # N(H+2)(W+2)C
        logits_d[:,:,0,:] = logits_d[:,:,1,:] # N(H+2)(W+2)C
        logits_d[:,:,-1,:] = logits_d[:,:,-2,:] # N(H+2)(W+2)C
        
        """
        | 4| 0| 5|
        | 2| 8| 3|
        | 6| 1| 7|
        """
        x_range = [1, -1,  0, 0, -1,  1, -1,  1, 0]
        y_range = [0,  0, -1, 1,  1,  1, -1, -1, 0]
        dist_maps = torch.zeros((0,len(x))).cuda() # 8k
        kl_maps = torch.zeros((0,len(x))).cuda() # 8k

        kl_center = logits[(n,x,y)] # KC

        for dx, dy in zip(x_range, y_range):
            dist_now = pred_dist_map_d[(n,x+dx+1,y+dy+1)]
            dist_maps = torch.cat((dist_maps,dist_now.unsqueeze(0)),0)

            if dx != 0 or dy != 0:
                logits_now = logits_d[(n,x+dx+1,y+dy+1)]
                # kl_map_now = torch.kl_div((kl_center+eps).log(), logits_now+eps).sum(2)  # 8KC->8K
                if self.isdetach:
                    logits_now = logits_now.detach()
                kl_map_now = kl_div(kl_center, logits_now)
                
                kl_map_now = kl_map_now.sum(1)  # KC->K
                # kl_maps = torch.cat((kl_maps,kl_map_now.unsqueeze(0)),0)
                kl_maps = torch.cat((kl_maps,kl_map_now.unsqueeze(0).cuda()),0)
                torch.clamp(kl_maps, min=0.0, max=20.0)

        # direction_gt shound be Nk  (8k->K)
        direction_gt = torch.argmin(dist_maps, dim=0)
        # weight_ce = pred_dist_map[bound]
        weight_ce = pred_dist_map[(n,x,y)]
        # print(weight_ce)

        # delete if min is 8 (local position)
        direction_gt_idx = [direction_gt!=8]
        direction_gt = direction_gt[direction_gt_idx]


        kl_maps = torch.transpose(kl_maps,0,1)
        direction_pred = kl_maps[direction_gt_idx]
        weight_ce = weight_ce[direction_gt_idx]

        return direction_gt, direction_pred, weight_ce

    def get_dist_maps(self, target):
        target_detach = target.clone().detach()
        dist_maps = torch.cat([self.dist_map_transform(target_detach[i]) for i in range(target_detach.shape[0])])
        out = -dist_maps
        out = torch.where(out>0, out, torch.zeros_like(out))
        
        return out

    def forward(self, logits, target):
        eps = 1e-10
        ph, pw = logits.size(2), logits.size(3)
        h, w = target.size(1), target.size(2)

        if ph != h or pw != w:
            logits = F.interpolate(input=logits, size=(
                h, w), mode='bilinear', align_corners=True)

        gt_boundary = self.gt2boundary(target, ignore_label=self.ignore_label)

        dist_maps = self.get_dist_maps(gt_boundary).cuda() # <-- it will slow down the training, you can put it to dataloader.

        pred_boundary = self.logits2boundary(logits)
        if pred_boundary.sum() < 1: # avoid nan
            return None # you should check in the outside. if None, skip this loss.
        
        direction_gt, direction_pred, weight_ce = self.get_direction_gt_predkl(dist_maps, pred_boundary, logits) # NHW,NHW,NCHW

        # direction_pred [K,8], direction_gt [K]
        loss = self.criterion(direction_pred, direction_gt) # careful
        
        weight_ce = self.weight_func(weight_ce)
        loss = (loss * weight_ce).mean()  # add distance weight

        return loss


if __name__ == '__main__':
    from torch.backends import cudnn
    import os
    import random
    cudnn.benchmark = False
    cudnn.deterministic = True

    seed = 0
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    random.seed(seed)
    np.random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)

    n,c,h,w = 1,2,100,100
    gt = torch.zeros((n,h,w)).cuda()
    gt[0,5] = 1
    gt[0,50] = 1
    logits = torch.randn((n,c,h,w)).cuda()

    abl = ABL()
    print(abl(logits, gt))


import numpy as np
import torch


def norm_angle(angle, angle_range):
    """Limit the range of angles.

    Args:
        angle (ndarray): shape(n, ).
        angle_range (Str): angle representations.

    Returns:
        angle (ndarray): shape(n, ).
    """
    if angle_range == 'oc':
        return angle
    elif angle_range == 'le135':
        return (angle + np.pi / 4) % np.pi - np.pi / 4
    elif angle_range == 'le90':
        return (angle + np.pi / 2) % np.pi - np.pi / 2
    else:
        print('Not yet implemented.')


def delta2theta(rois, deltas, rois_mode='hbbox'):
    """Based on 
                delta2bbox from mmrotate/core/bbox/coder/delta_xywha_hbbox_coder.py:

    def __init__(self,
                 target_means=(0., 0., 0., 0., 0.),
                 target_stds=(1., 1., 1., 1., 1.),
                 angle_range='oc',
                 norm_factor=None,
                 edge_swap=False,
                 clip_border=True,
                 add_ctr_clamp=False,
                 ctr_clamp=32):

    ------- and configs/rotated_faster_rcnn/rotated_faster_rcnn_r50_fpn_1x_dota_le90.py:

            bbox_coder=dict(
                type='DeltaXYWHAHBBoxCoder',
                angle_range=angle_version,
                norm_factor=2,
                edge_swap=True,
                target_means=(.0, .0, .0, .0, .0),
                target_stds=(0.1, 0.1, 0.2, 0.2, 0.1)),

    ------- and delta2bbox from mmrotate/core/bbox/coder/delta_xywha_rbbox_coder.py:

    def __init__(self,
                 target_means=(0., 0., 0., 0., 0.),
                 target_stds=(1., 1., 1., 1., 1.),
                 angle_range='oc',
                 norm_factor=None,
                 edge_swap=False,
                 proj_xy=False,
                 add_ctr_clamp=False,
                 ctr_clamp=32):

    ------- and configs/oriented_rcnn/oriented_rcnn_r50_fpn_1x_dota_le90.py:

            bbox_coder=dict(
                type='DeltaXYWHAOBBoxCoder',
                angle_range=angle_version,
                norm_factor=None,
                edge_swap=True,
                proj_xy=True,
                target_means=(.0, .0, .0, .0, .0),
                target_stds=(0.1, 0.1, 0.2, 0.2, 0.1)),
    """

    if rois_mode == 'hbbox':
        means=(.0, .0, .0, .0, .0)
        stds=(0.1, 0.1, 0.2, 0.2, 0.1)
        # max_shape=None
        wh_ratio_clip=16 / 1000
        add_ctr_clamp=False
        ctr_clamp=32
        angle_range='le90'
        norm_factor=2
        # edge_swap=True
        # proj_xy=False
    elif rois_mode == 'rbbox':
        means=(.0, .0, .0, .0, .0)
        stds=(0.1, 0.1, 0.2, 0.2, 0.1)
        # max_shape=None
        wh_ratio_clip=16 / 1000
        add_ctr_clamp=False
        ctr_clamp=32
        angle_range='le90'
        norm_factor=None
        # edge_swap=True
        # proj_xy=True
    else:
        raise NotImplementedError

    means = deltas.new_tensor(means).view(1, -1).repeat(1, deltas.size(1) // 5)
    stds = deltas.new_tensor(stds).view(1, -1).repeat(1, deltas.size(1) // 5)
    denorm_deltas = deltas * stds + means
    dx = denorm_deltas[:, 0::5]
    dy = denorm_deltas[:, 1::5]
    dw = denorm_deltas[:, 2::5]
    dh = denorm_deltas[:, 3::5]
    da = denorm_deltas[:, 4::5]
    if norm_factor:
        da *= norm_factor * np.pi

    if rois_mode == 'hbbox':
        x1, y1, x2, y2 = rois.unbind(dim=-1)
        # Compute center of each roi
        px = ((x1 + x2) * 0.5).unsqueeze(-1).expand_as(dx)
        py = ((y1 + y2) * 0.5).unsqueeze(-1).expand_as(dy)
        # Compute width/height of each roi
        pw = (x2 - x1).unsqueeze(-1).expand_as(dw)
        ph = (y2 - y1).unsqueeze(-1).expand_as(dh)
    elif rois_mode == 'rbbox':
        # Compute center of each roi
        px = rois[:, 0].unsqueeze(1).expand_as(dx)
        py = rois[:, 1].unsqueeze(1).expand_as(dy)
        # Compute width/height of each roi
        pw = rois[:, 2].unsqueeze(1).expand_as(dw)
        ph = rois[:, 3].unsqueeze(1).expand_as(dh)
        # Compute rotated angle of each roi
        # pa = rois[:, 4].unsqueeze(1).expand_as(da)
    else:
        pass

    dx_width = pw * dx
    dy_height = ph * dy

    max_ratio = np.abs(np.log(wh_ratio_clip))
    if add_ctr_clamp:
        dx_width = torch.clamp(dx_width, max=ctr_clamp, min=-ctr_clamp)
        dy_height = torch.clamp(dy_height, max=ctr_clamp, min=-ctr_clamp)
        dw = torch.clamp(dw, max=max_ratio)
        dh = torch.clamp(dh, max=max_ratio)
    else:
        dw = dw.clamp(min=-max_ratio, max=max_ratio)
        dh = dh.clamp(min=-max_ratio, max=max_ratio)

    ga = norm_angle(da, angle_range)

    a11 =  ga.cos() / dw.exp()
    a12 =  ga.sin() / dw.exp() * ph / (pw + 1e-5)
    a21 = -ga.sin() / dh.exp() * pw / (ph + 1e-5)
    a22 =  ga.cos() / dh.exp()
    a13 = -2 * dx * a11 - 2 * dy * a12
    a23 = -2 * dx * a21 - 2 * dy * a22

    return torch.cat([a11, a12, a13, a21, a22, a23], dim=-1)


# ===== ===== ===== ===== ===== ===== ===== ===== ===== =====


# 推导过程
# 
# 
# 我们试图找到一种将[dx, dy, dw, dh, da]转换为
# [[a11, a12, a13],
#  [a21, a22, a23]]
# 仿射矩阵的方法
# 因为只要找到仿射矩阵，就能创建出一个0-1 activation mask去引导attention
# 
# 注意到
# (1)
#     denorm_deltas = deltas * stds + means
#     ...
#     if norm_factor:
#         da *= norm_factor * np.pi
# (2)
#     gx = px + pw * dx
#     gy = py + ph * dy
#     gw = pw * dw.exp()
#     gh = ph * dh.exp()
#     ga = norm_angle(da, angle_range)
# 
# 因此仿射矩阵应该是
# [[x'], = M[RoI→7x7] @ M[a] @ M[wh] @ M[7x7→RoI] @ [[x], + M[xy]
#  [y']]                                             [y]]
# （先回到RoI状态，再缩放，再旋转，再回到7x7特征图，最后平移）
# 其中
# M[7x7→RoI] = [[pw / 2  , 0        ],
#               [0       , ph / 2   ]]
# M[wh]      = [[dw.exp(), 0        ],
#               [0       , dh.exp() ]]
# M[a]       = [[ga.cos(), -ga.sin()],
#               [ga.sin(),  ga.cos()]]
# M[RoI→7x7] = [[2 / pw  , 0        ],
#               [0       , 2 / ph   ]]
# M[xy]      = [[2 * dx],
#               [2 * dy]]
# (这是由于grid会将output视为[-1, 1]归一化的图）
# 
# 值得注意的是
# 我们的坐标系是左手平面直角坐标系
# 因此对于xywha长边定义法中的a，顺时针为正，逆时针为负（即x轴向y轴转记为正）
# 而旋转矩阵仍然相同，例如：
# 在左手平面直角坐标系中，从坐标(-1, -1)旋转+90°到(1, -1)满足
# [[ 1], = [[0, -1], @ [[-1],
#  [-1]]    [1,  0]]    [-1]]
# 这一点也可以参考https://zhuanlan.zhihu.com/p/459018810
# 
# 因此
# [[x'], = [[ga.cos() * dw.exp()          , -ga.sin() * dh.exp() * ph / pw], @ [[x], + [[2 * dx],
#  [y']]    [ga.sin() * dw.exp() * pw / ph,  ga.cos() * dh.exp()          ]]    [y]]    [2 * dy]]
# （对角矩阵在前横乘，对角矩阵在后竖乘）
# 
# 在应用F.affine_grid(theta, size) & F.grid_sample(input, grid)函数对时
# 应当注意到
# (1) F.affine_grid(theta, size)是对一个大小为size=N×C×H×W的张量的每个格点坐标应用仿射矩阵theta、给出每个格点的新坐标
# (2) F.grid_sample(input, grid)是对input应用grid采样，得到大小为size=N×C×H×W的张量output
# 所以实际上是
# input = theta(output)
# output = theta^-1(input)
#    
# 对于二阶矩阵
# A = [[a, b],
#      [c, d]]
# 其逆矩阵
# A^-1 = 1 / (ad - bc) * [[ d, -b],
#                         [-c,  a]]
# 因此
# [[a11, a12], = 1 / (dw.exp() * dh.exp()) * [[ ga.cos() * dh.exp()          , ga.sin() * dh.exp() * ph / pw],
#  [a21, a22]]                                [-ga.sin() * dw.exp() * pw / ph, ga.cos() * dw.exp()          ]]
#              =                             [[ ga.cos() / dw.exp()          , ga.sin() / dw.exp() * ph / pw],
#                                             [-ga.sin() / dh.exp() * pw / ph, ga.cos() / dh.exp()          ]]
# [[a13], = [[a11, a12], @ [[-2 * dx],
#  [a23]]    [a21, a22]]    [-2 * dy]]
#         = [[-2 * dx * a11 - 2 * dy * a12],
#            [-2 * dx * a21 - 2 * dy * a22]]


# ===== ===== ===== ===== ===== ===== ===== ===== ===== =====


# 验证过程


# 假设pw = 200, ph = 100, dw.exp() = 0.5, dh.exp() = 0.5, dx = 1/7, dy = -1/7, ga = da = np.pi / 4
# 则根据上式可计算出
# [[a11, a12, a13], = [[ 1.4142,  0.7071,  1.4142 * (-2/7) + 0.7071 * (+2/7)],
#  [a21, a22, a23]]    [-2.8284,  1.4142, -2.8284 * (-2/7) + 1.4142 * (+2/7)]]

# 验证代码：
#    import torch
#    import torch.nn.functional as F

#    #x = torch.ones(1, 1, 5, 5)

#    #theta_o = torch.tensor([
#    #    [[ 1.0000,  0.0000,  0.0000],
#    #     [ 0.0000,  1.0000,  0.0000]]
#    #    ], dtype=torch.float)
#    #theta_x = torch.tensor([
#    #    [[ 1.0000,  0.0000,  0.5000],
#    #     [ 0.0000,  1.0000,  0.0000]]
#    #    ], dtype=torch.float)
#    #theta_y = torch.tensor([
#    #    [[ 1.0000,  0.0000,  0.0000],
#    #     [ 0.0000,  1.0000,  0.5000]]
#    #    ], dtype=torch.float)
#    #theta_w = torch.tensor([
#    #    [[ 2.0000,  0.0000,  0.0000],
#    #     [ 0.0000,  1.0000,  0.0000]]
#    #    ], dtype=torch.float)
#    #theta_h = torch.tensor([
#    #    [[ 1.0000,  0.0000,  0.0000],
#    #     [ 0.0000,  2.0000,  0.0000]]
#    #    ], dtype=torch.float)
#    #theta_a = torch.tensor([
#    #    [[ 0.0000, -1.0000,  0.0000],
#    #     [ 1.0000,  0.0000,  0.0000]]
#    #    ], dtype=torch.float)

#    #grid_o = F.affine_grid(theta_o, x.size())
#    #print(grid_o)
#    #grid_x = F.affine_grid(theta_x, x.size())
#    #print(grid_x)
#    #grid_y = F.affine_grid(theta_y, x.size())
#    #print(grid_y)
#    #grid_w = F.affine_grid(theta_w, x.size())
#    #print(grid_w)
#    #grid_h = F.affine_grid(theta_h, x.size())
#    #print(grid_h)
#    #grid_a = F.affine_grid(theta_a, x.size())
#    #print(grid_a)

#    x = torch.ones(1, 1, 7, 7)

#    theta = torch.tensor([
#        [[ 1.4142,  0.7071,  1.4142 * (-2/7) + 0.7071 * (+2/7)],
#         [-2.8284,  1.4142, -2.8284 * (-2/7) + 1.4142 * (+2/7)]]
#        ], dtype=torch.float)

#    grid = F.affine_grid(theta, x.size())
#    x = F.grid_sample(x, grid)
#    print(x)

# 输出结果：
#    tensor([[[[0.0000, 0.0000, 0.0000, 1.0000, 1.0000, 0.0000, 0.0000],
#              [0.0000, 0.0000, 0.0000, 1.0000, 1.0000, 0.0000, 0.0000],
#              [0.0000, 0.0000, 0.0000, 1.0000, 1.0000, 1.0000, 0.0000],
#              [0.0000, 0.0000, 0.0000, 0.0000, 1.0000, 1.0000, 0.0000],
#              [0.0000, 0.0000, 0.0000, 0.0000, 1.0000, 1.0000, 0.0000],
#              [0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.4645, 0.0000],
#              [0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000]]]])

# 与我们的预期是比较符合的。


# ===== ===== ===== ===== ===== ===== ===== ===== ===== =====




import glob 
import os
import os.path as osp
from ultralytics import YOLO

import imgviz
import json
import pandas as pd
from tqdm import tqdm
import numpy as np

from visionsuite.utils.visualizers.vis_obb import vis_obb


input_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/datasets/split_labelme_patch_dataset_yolo_iseg/images/val'
weights_file = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/outputs/yolo/train2/weights/best.pt'
_classes = ['SCREW', 'TIMBER']
output_dir = "/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/outputs/yolo/tests"
input_img_ext = 'jpg'
imgsz = 1024
iou_threshold = 0.8
conf_threshold = 0.25

model = YOLO(weights_file) 

if not osp.exists(output_dir):
    os.makedirs(output_dir)

img_files = glob.glob(osp.join(input_dir, f'*.{input_img_ext}'))

for img_file in tqdm(img_files):
    model(img_file, save=True, imgsz=imgsz, conf=conf_threshold, verbose=True, device='1')
    


# _idx2class = {idx: cls for idx, cls in enumerate(_classes)}

# if not osp.exists(output_dir):
#     os.makedirs(output_dir)

# img_files = glob.glob(osp.join(input_dir, f'*.{input_img_ext}'))

# preds = {}
# compare = {}
# for img_file in tqdm(img_files):
#     filename = osp.split(osp.splitext(img_file)[0])[-1]
#     pred = model(img_file, save=False, imgsz=imgsz, iou=iou_threshold, conf=conf_threshold, verbose=False)[0]
    
#     idx2class = pred.names
#     obb_result = pred.obb
#     classes = obb_result.cls.tolist()
#     confs = obb_result.conf.tolist()
    
#     for res in obb_result:
#         xywhr = res.xywhr
#         angle = np.rad2deg(xywhr[0][-1].cpu().detach().item())
#         # print("angle: ", angle)
#         xyxyxyxy = res.xyxyxyxy
#         # print("xyxyxyxyxy: ", xyxyxyxy)
    
#     idx2xyxys = {}
#     for cls, xyxys, conf in zip(classes, obb_result.xyxyxyxy, confs):
#         if cls not in idx2xyxys:
#             idx2xyxys[cls] = {'polygon': [], 'confidence': []}
#         idx2xyxys[cls]['polygon'].append([xy.tolist() for xy in xyxys])
#         idx2xyxys[cls]['confidence'].append(conf)
    
#     if _classes is not None:
#         new_idx2xyxys = {}
#         for idx, _cls in enumerate(_classes):
#             for jdx, cls in enumerate(idx2class.values()):
#                 if cls == _cls and jdx in idx2xyxys:
#                     new_idx2xyxys[idx] = idx2xyxys[jdx]
        
#         idx2xyxys = new_idx2xyxys
#         idx2class = _idx2class   
    
#     color_map = imgviz.label_colormap()[1:len(idx2class) + 1 + 1]
#     preds.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
    
#     if compare_gt:
#         _compare = vis_obb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
#                             compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale,
#                             output_img_ext=output_img_ext, output_img_size_ratio=output_img_size_ratio)
#         _compare.update({"img_file": img_file})
#         compare.update({filename: _compare})
#     else:
#         vis_obb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
#                 compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale,
#                 output_img_ext=output_img_ext, output_img_size_ratio=output_img_size_ratio)
            
# with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
#     json.dump(preds, json_file, ensure_ascii=False, indent=4)

# if compare_gt:
#     with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
#         json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
#     df_compare = pd.DataFrame(compare)
#     df_compare_pixel = df_compare.loc['diff_iou'].T
#     df_compare_pixel.fillna(0, inplace=True)
#     df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp
import time
import warnings
import numpy as np
import imgviz
import json
import mmcv
import pandas as pd
import torch
from mmcv import Config, DictAction
from mmcv.cnn import fuse_conv_bn
from mmcv.parallel import MMDataParallel, MMDistributedDataParallel
from mmcv.runner import (get_dist_info, init_dist, load_checkpoint,
                         wrap_fp16_model)
from mmdet.apis import multi_gpu_test#, single_gpu_test
from mmdet.datasets import build_dataloader, replace_ImageToTensor

from mmrotate.datasets import build_dataset
from mmrotate.models import build_detector
from mmrotate.utils import compat_cfg, setup_multi_processes
import visionsuite.cores.spatial_transform_decoupling.mmrotate_custom.datasets.custom_dota_dataset as custom_dota_dataset
from visionsuite.utils.visualizers.vis_obb import vis_obb
from visionsuite.utils.helpers import JsonEncoder



def parse_args():
    """Parse parameters."""
    parser = argparse.ArgumentParser(
        description='MMDet test (and eval) a model')
    parser.add_argument('--config', default='/HDD/etc/std/Spatial-Transform-Decoupling/src/configs/rotated_imted/dota/vit/rotated_imted_vb1m_oriented_rcnn_vit_base_1x_dota_ms_rr_le90_stdc_xyawh321v.py')
    parser.add_argument('--checkpoint', default='/HDD/_projects/github/VisionSuite/visionsuite/cores/spatial_transform_decoupling/work_dirs/rotated_imted_vb1m_oriented_rcnn_vit_base_1x_dota_ms_rr_le90_stdc_xyawh321v/epoch_300.pth')
    parser.add_argument(
        '--work-dir',
        help='the directory to save the file containing evaluation metrics')
    parser.add_argument('--out', help='output result file in pickle format')
    parser.add_argument(
        '--fuse-conv-bn',
        action='store_true',
        help='Whether to fuse conv and bn, this will slightly increase'
        'the inference speed')
    parser.add_argument(
        '--gpu-ids',
        default=[0, 1],
        type=int,
        nargs='+',
        help='ids of gpus to use '
        '(only applicable to non-distributed testing)')
    parser.add_argument(
        '--format-only',
        action='store_true',
        help='Format the output results without perform evaluation. It is'
        'useful when you want to format the result to a specific format and '
        'submit it to the test server')
    parser.add_argument(
        '--eval',
        type=str,
        nargs='+',
        help='evaluation metrics, which depends on the dataset, e.g., "bbox",'
        ' "segm", "proposal" for COCO, and "mAP", "recall" for PASCAL VOC')
    parser.add_argument('--show', action='store_true', help='show results')
    parser.add_argument(
        '--show-dir', default='/HDD/etc/std/test_300')
    parser.add_argument(
        '--show-score-thr',
        type=float,
        default=0.3,
        help='score threshold (default: 0.3)')
    parser.add_argument(
        '--gpu-collect',
        action='store_true',
        help='whether to use gpu to collect results.')
    parser.add_argument(
        '--tmpdir',
        help='tmp directory used for collecting results from multiple '
        'workers, available when gpu-collect is not specified')
    parser.add_argument(
        '--cfg-options',
        nargs='+',
        action=DictAction,
        help='override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. If the value to '
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        'Note that the quotation marks are necessary and that no white space '
        'is allowed.')
    parser.add_argument(
        '--eval-options',
        nargs='+',
        action=DictAction,
        help='custom options for evaluation, the key-value pair in xxx=yyy '
        'format will be kwargs for dataset.evaluate() function')
    parser.add_argument(
        '--launcher',
        choices=['none', 'pytorch', 'slurm', 'mpi'],
        default='none',
        help='job launcher')
    parser.add_argument('--local_rank', type=int, default=0)
    args = parser.parse_args()
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = str(args.local_rank)

    return args


def main():
    # ~/anaconda3/envs/openmmlab/lib/python3.7/site-packages/mmdet/models/dense_heads/anchor_head.py:123: UserWarning: 
    # DeprecationWarning: anchor_generator is deprecated, please use "prior_generator" instead
    #   warnings.warn('DeprecationWarning: anchor_generator is deprecated, '
    warnings.filterwarnings("ignore", category=UserWarning)

    args = parse_args()

    assert args.out or args.eval or args.format_only or args.show \
        or args.show_dir, \
        ('Please specify at least one operation (save/eval/format/show the '
         'results / save the results) with the argument "--out", "--eval"'
         ', "--format-only", "--show" or "--show-dir"')

    if args.eval and args.format_only:
        raise ValueError('--eval and --format_only cannot be both specified')

    if args.out is not None and not args.out.endswith(('.pkl', '.pickle')):
        raise ValueError('The output file must be a pkl file.')

    cfg = Config.fromfile(args.config)
    if args.cfg_options is not None:
        cfg.merge_from_dict(args.cfg_options)

    cfg = compat_cfg(cfg)

    # set multi-process settings
    setup_multi_processes(cfg)

    # set cudnn_benchmark
    if cfg.get('cudnn_benchmark', False):
        torch.backends.cudnn.benchmark = True

    cfg.model.pretrained = None
    if cfg.model.get('neck'):
        if isinstance(cfg.model.neck, list):
            for neck_cfg in cfg.model.neck:
                if neck_cfg.get('rfp_backbone'):
                    if neck_cfg.rfp_backbone.get('pretrained'):
                        neck_cfg.rfp_backbone.pretrained = None
        elif cfg.model.neck.get('rfp_backbone'):
            if cfg.model.neck.rfp_backbone.get('pretrained'):
                cfg.model.neck.rfp_backbone.pretrained = None

    if args.gpu_ids is not None:
        cfg.gpu_ids = args.gpu_ids
    else:
        cfg.gpu_ids = range(1)

    # init distributed env first, since logger depends on the dist info.
    if args.launcher == 'none':
        distributed = False
        if len(cfg.gpu_ids) > 1:
            warnings.warn(
                f'We treat {cfg.gpu_ids} as gpu-ids, and reset to '
                f'{cfg.gpu_ids[0:1]} as gpu-ids to avoid potential error in '
                'non-distribute testing time.')
            cfg.gpu_ids = cfg.gpu_ids[0:1]
    else:
        distributed = True
        init_dist(args.launcher, **cfg.dist_params)

    test_dataloader_default_args = dict(
        samples_per_gpu=1, workers_per_gpu=2, dist=distributed, shuffle=False)

    # in case the test dataset is concatenated
    if isinstance(cfg.data.test, dict):
        cfg.data.test.test_mode = True
        if 'samples_per_gpu' in cfg.data.test:
            warnings.warn('`samples_per_gpu` in `test` field of '
                          'data will be deprecated, you should'
                          ' move it to `test_dataloader` field')
            test_dataloader_default_args['samples_per_gpu'] = \
                cfg.data.test.pop('samples_per_gpu')
        if test_dataloader_default_args['samples_per_gpu'] > 1:
            # Replace 'ImageToTensor' to 'DefaultFormatBundle'
            cfg.data.test.pipeline = replace_ImageToTensor(
                cfg.data.test.pipeline)
    elif isinstance(cfg.data.test, list):
        for ds_cfg in cfg.data.test:
            ds_cfg.test_mode = True
            if 'samples_per_gpu' in ds_cfg:
                warnings.warn('`samples_per_gpu` in `test` field of '
                              'data will be deprecated, you should'
                              ' move it to `test_dataloader` field')
        samples_per_gpu = max(
            [ds_cfg.pop('samples_per_gpu', 1) for ds_cfg in cfg.data.test])
        test_dataloader_default_args['samples_per_gpu'] = samples_per_gpu
        if samples_per_gpu > 1:
            for ds_cfg in cfg.data.test:
                ds_cfg.pipeline = replace_ImageToTensor(ds_cfg.pipeline)

    test_loader_cfg = {
        **test_dataloader_default_args,
        **cfg.data.get('test_dataloader', {})
    }

    rank, _ = get_dist_info()
    # allows not to create
    if args.work_dir is not None and rank == 0:
        mmcv.mkdir_or_exist(osp.abspath(args.work_dir))
        timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
        json_file = osp.join(args.work_dir, f'eval_{timestamp}.json')

    # build the dataloader
    dataset = build_dataset(cfg.data.test)
    data_loader = build_dataloader(dataset, **test_loader_cfg)

    # build the model and load checkpoint
    cfg.model.train_cfg = None
    model = build_detector(cfg.model, test_cfg=cfg.get('test_cfg'))
    fp16_cfg = cfg.get('fp16', None)
    if fp16_cfg is not None:
        wrap_fp16_model(model)
    checkpoint = load_checkpoint(model, args.checkpoint, map_location='cpu')
    if args.fuse_conv_bn:
        model = fuse_conv_bn(model)
    # old versions did not save class info in checkpoints, this walkaround is
    # for backward compatibility
    if 'CLASSES' in checkpoint.get('meta', {}):
        model.CLASSES = checkpoint['meta']['CLASSES']
    else:
        model.CLASSES = dataset.CLASSES

    model = MMDataParallel(model, device_ids=cfg.gpu_ids)

    class2idx = {'BOX': 0}
    idx2class = {0: 'BOX'}
    output_dir = '/HDD/_projects/benchmark/obb_detection/rich/tests/std_hivit'
    json_dir = '/HDD/_projects/benchmark/obb_detection/rich/datasets/split_dataset_box/val'
    iou_threshold = 0.2
    
    if not osp.exists(output_dir):
        os.mkdir(output_dir)
    compare_gt = True
    color_map = imgviz.label_colormap()[1:len(idx2class) + 1 + 1]

    model.eval()
    dataset = data_loader.dataset
    idx2xyxys = {}
    compare = {}
    results = {}
    for i, data in enumerate(data_loader):
        
        img_metas = data['img_metas'][0].data[0]
        with torch.no_grad():
            result = model(return_loss=False, rescale=True, **data)
            
        for idx, res in enumerate(result):
            labels = [
                np.full(bbox[0].shape[0], i, dtype=int)
                for i, bbox in enumerate([res])
            ]
            labels = np.concatenate(labels)
            bboxes = np.vstack(res)
            
            _idx2xyxys = {}
            for i, (label, bbox) in enumerate(zip(labels, bboxes)):
                
                xc, yc, w, h, ag = bbox[:5]
                wx, wy = w / 2 * np.cos(ag), w / 2 * np.sin(ag)
                hx, hy = -h / 2 * np.sin(ag), h / 2 * np.cos(ag)
                p1 = [xc - wx - hx, yc - wy - hy]
                p2 = [xc + wx - hx, yc + wy - hy]
                p3 = [xc + wx + hx, yc + wy + hy]
                p4 = [xc - wx + hx, yc - wy + hy]
                poly = [p1, p2, p3, p4]
                
                if label not in _idx2xyxys:
                    _idx2xyxys[int(label)] = {'polygon': [poly], 'confidence': [bbox[-1]]}
                else:
                    _idx2xyxys[int(label)]['polygon'].append(poly)
                    _idx2xyxys[int(label)]['confidence'].append(bbox[-1])
                    
            
        img_file = img_metas[idx]['filename']
        filename = osp.split(osp.splitext(img_file)[0])[-1]
        results.update({filename: {'idx2xyxys': _idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
            
        _compare = vis_obb(img_file, _idx2xyxys, idx2class, output_dir, color_map, json_dir, 
                           compare_gt=compare_gt, iou_threshold=iou_threshold)
        _compare.update({"img_file": img_file})
        compare.update({filename: _compare})


    with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
        json.dump(results, json_file, ensure_ascii=False, indent=4, cls=JsonEncoder)

    if compare_gt:
        with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
            json.dump(compare, json_file, ensure_ascii=False, indent=4)
        
        df_compare = pd.DataFrame(compare)
        df_compare_pixel = df_compare.loc['diff_iou'].T
        df_compare_pixel.fillna(0, inplace=True)
        df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))

    # if not distributed:
    #     model = MMDataParallel(model, device_ids=cfg.gpu_ids)
    #     outputs = single_gpu_test(model, data_loader, args.show, args.show_dir,
    #                               args.show_score_thr)
    # else:
    #     model = MMDistributedDataParallel(
    #         model.cuda(),
    #         device_ids=[torch.cuda.current_device()],
    #         broadcast_buffers=False)
    #     outputs = multi_gpu_test(model, data_loader, args.tmpdir,
    #                              args.gpu_collect)

    # rank, _ = get_dist_info()
    # if rank == 0:
    #     if args.out:
    #         print(f'\nwriting results to {args.out}')
    #         mmcv.dump(outputs, args.out)
    #     kwargs = {} if args.eval_options is None else args.eval_options
    #     if args.format_only:
    #         dataset.format_results(outputs, **kwargs)
    #     if args.eval:
    #         eval_kwargs = cfg.get('evaluation', {}).copy()
    #         # hard-code way to remove EvalHook args
    #         for key in [
    #                 'interval', 'tmpdir', 'start', 'gpu_collect', 'save_best',
    #                 'rule', 'dynamic_intervals'
    #         ]:
    #             eval_kwargs.pop(key, None)
    #         eval_kwargs.update(dict(metric=args.eval, **kwargs))
    #         metric = dataset.evaluate(outputs, **eval_kwargs)
    #         print(metric)
    #         metric_dict = dict(config=args.config, metric=metric)
    #         if args.work_dir is not None and rank == 0:
    #             mmcv.dump(metric_dict, json_file)


if __name__ == '__main__':
    main()



import visionsuite.engines.classification.src # to pre-compile
from visionsuite.engines.utils.bases import BaseEngine
from visionsuite.engines.classification.utils.registry import RUNNERS

class Engine(BaseEngine):
    def __init__(self, task):
        super().__init__(task=task)
        
    def train(self, cfgs_file=None, *args, **kwargs):
        
        # TODO: need to assign by yaml file
        runner = RUNNERS.get("TrainRunner", case_sensitive=True)('classification')
        assert runner is not None, ValueError(f"runner is None")
        
        runner.train(cfgs_file=cfgs_file, *args, **kwargs)
                        
        
        


########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################


from cityscapesscripts.helpers.labels import labels


class CityscapesBase:
    SPLITS = ["train", "valid", "test", "val"]

    # number of classes without void/unlabeled and license plate (class 34)
    N_CLASSES = [19, 33]

    # 1+33 classes (0: unlabeled)
    CLASS_NAMES_FULL = [label.name for label in labels]
    CLASS_COLORS_FULL = [label.color for label in labels]

    # 1+19 classes (0: void)
    CLASS_NAMES_REDUCED = ["void"] + [
        label.name for label in labels if not label.ignoreInEval
    ]
    CLASS_COLORS_REDUCED = [(0, 0, 0)] + [
        label.color for label in labels if not label.ignoreInEval
    ]
    # forward mapping (0: unlabeled) + 33 classes -> (0: void) + 19 classes
    CLASS_MAPPING_REDUCED = {
        c: labels[c].trainId + 1 if not labels[c].ignoreInEval else 0
        for c in range(1 + 33)
    }

    RGB_DIR = "rgb"

    LABELS_FULL_DIR = "labels_33"
    LABELS_FULL_COLORED_DIR = "labels_33_colored"

    LABELS_REDUCED_DIR = "labels_19"
    LABELS_REDUCED_COLORED_DIR = "labels_19_colored"



# Copyright (c) ByteDance, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import os
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as checkpoint

from functools import partial
from visionsuite.cores.spatial_transform_decoupling.mmcv_custom import load_checkpoint_hivit_1 as load_checkpoint
from mmdet.utils import get_root_logger
from mmdet.models.builder import BACKBONES
import visionsuite.cores.spatial_transform_decoupling.models as models


def window_partition(x, window_size):
    """
    Partition into non-overlapping windows with padding if needed.
    Args:
        x (tensor): input tokens with [B, H, W, C].
        window_size (int): window size.

    Returns:
        windows: windows after partition with [B * num_windows, window_size, window_size, C].
        (Hp, Wp): padded height and width before partition
    """
    B, H, W, C = x.shape

    pad_h = (window_size - H % window_size) % window_size
    pad_w = (window_size - W % window_size) % window_size
    if pad_h > 0 or pad_w > 0:
        x = F.pad(x, (0, 0, 0, pad_w, 0, pad_h))
    Hp, Wp = H + pad_h, W + pad_w

    x = x.view(B, Hp // window_size, window_size, Wp // window_size, window_size, C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    return windows, (Hp, Wp)


def window_unpartition(windows, window_size, pad_hw, hw):
    """
    Window unpartition into original sequences and removing padding.
    Args:
        x (tensor): input tokens with [B * num_windows, window_size, window_size, C].
        window_size (int): window size.
        pad_hw (Tuple): padded height and width (Hp, Wp).
        hw (Tuple): original height and width (H, W) before padding.

    Returns:
        x: unpartitioned sequences with [B, H, W, C].
    """
    Hp, Wp = pad_hw
    H, W = hw
    B = windows.shape[0] // (Hp * Wp // window_size // window_size)
    x = windows.view(B, Hp // window_size, Wp // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, Hp, Wp, -1)

    if Hp > H or Wp > W:
        x = x[:, :H, :W, :].contiguous()
    return x


class LayerNorm2d(nn.LayerNorm):
    def forward(self, x):
        x = x.permute(0, 2, 3, 1).contiguous()
        x = super().forward(x)
        x = x.permute(0, 3, 1, 2).contiguous()
        return x


@BACKBONES.register_module()
class HiViT(models.HiViT):
    def __init__(self,
                 img_size=224,
                 patch_size=16,
                 in_chans=3,
                 embed_dim=512,
                 depths=[4, 4, 20], 
                 num_heads=8,
                 stem_mlp_ratio=3., 
                 mlp_ratio=4.,
                 drop_path_rate=0.1,
                 norm_layer=partial(nn.LayerNorm, eps=1e-6), 
                 ape=True, rpe=True, 
                 patch_norm=True,
                 with_fpn=False,
                 with_simple_fpn=False,
                 out_indices=[4, 9, 14, 19],
                 use_checkpoint=False,
                 global_indices=[4, 9, 14, 19], 
                 window_size=None,
                 init_cfg=None,
                 last_feat=False,
                 **kwargs):
        super(HiViT, self).__init__(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
            depths=depths,
            num_heads=num_heads,
            stem_mlp_ratio=stem_mlp_ratio,
            mlp_ratio=mlp_ratio,
            drop_path_rate=drop_path_rate,
            norm_layer=norm_layer,
            ape=ape, rpe=rpe,
            patch_norm=patch_norm,
            **kwargs)

        assert not (with_fpn or with_simple_fpn) or (patch_size in (16,))
        assert not (with_fpn and with_simple_fpn)
        self.init_cfg = init_cfg
        self.patch_size = patch_size
        self.with_fpn = with_fpn
        self.with_simple_fpn = with_simple_fpn
        self.merge_indices = (depths[0] * 2 - 1, depths[0] * 2 + depths[1] * 2)
        self.out_indices = out_indices
        self.use_checkpoint = use_checkpoint
        self.global_indices = global_indices
        self.window_size = window_size

        self.last_feat = last_feat

        del self.fc_norm, self.head, self.num_classes
        if with_fpn:
            self.fpn1 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
                # nn.BatchNorm2d(embed_dim),
                nn.GELU(),
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            ) if 'H' not in out_indices else LayerNorm2d(embed_dim // 4, eps=1e-6)
            self.fpn2 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            ) if 'M' not in out_indices else LayerNorm2d(embed_dim // 2, eps=1e-6)
            self.fpn3 = LayerNorm2d(embed_dim, eps=1e-6)
            self.fpn4 = nn.Sequential(
                nn.MaxPool2d(kernel_size=2, stride=2),
                LayerNorm2d(embed_dim, eps=1e-6),
            )
        else:
            logger = get_root_logger()
            logger.info('Build model without FPN.')

    def init_weights(self):
        if self.init_cfg is None:
            raise ValueError
        else:
            assert 'checkpoint' in self.init_cfg, f'Only support ' \
                                                  f'specify `Pretrained` in ' \
                                                  f'`init_cfg` in ' \
                                                  f'{self.__class__.__name__} '
            self.apply(self._init_weights)
            pretrained = self.init_cfg['checkpoint']
            logger = get_root_logger()
            if os.path.isfile(pretrained):
                load_checkpoint(self, pretrained, strict=False, logger=logger)
            else:
                raise ValueError(f"checkpoint path {pretrained} is invalid")

    def interpolate_pos_encoding(self, x, h, w):
        npatch = x.shape[1]
        N = self.absolute_pos_embed.shape[1]
        if npatch == N and w == h:
            return self.absolute_pos_embed
        patch_pos_embed = self.absolute_pos_embed
        dim = x.shape[-1]
        w0 = w // self.patch_size
        h0 = h // self.patch_size
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        w0, h0 = w0 + 0.1, h0 + 0.1
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
            scale_factor=(h0 / math.sqrt(N), w0 / math.sqrt(N)),
            mode='bicubic',
        )
        assert int(h0) == patch_pos_embed.shape[-2] and int(w0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return patch_pos_embed

    def prepare_tokens(self, x, mask=None):
        B, _, H, W = x.shape
        Hp, Wp = H // self.patch_size, W // self.patch_size
        features = []

        x = self.patch_embed(x)
        for i, blk in enumerate(self.blocks[:-self.num_main_blocks]):
            x = checkpoint.checkpoint(blk, x) if self.use_checkpoint else blk(x)
            if i == self.merge_indices[0] and 'H' in self.out_indices:
                xp = x.reshape(B, Hp, Wp, 4, 4, -1).permute(
                    0, 5, 1, 3, 2, 4
                ).reshape(B, -1, Hp*4, Wp*4).contiguous()
                for _ in range(self.out_indices.count('H')):
                    features.append(xp)
            if i == self.merge_indices[1] and 'M' in self.out_indices:
                xp = x.reshape(B, Hp, Wp, 2, 2, -1).permute(
                    0, 5, 1, 3, 2, 4
                ).reshape(B, -1, Hp*2, Wp*2).contiguous()
                for _ in range(self.out_indices.count('M')):
                    features.append(xp)
        x = x[..., 0, 0, :]
        # add positional encoding to each token
        x = x + self.interpolate_pos_encoding(x, H, W)
        return self.pos_drop(x), features
    
    def forward(self, x):
        B, _, H, W = x.shape
        Hp, Wp = H // self.patch_size, W // self.patch_size
        x, features = self.prepare_tokens(x)
        rpe_index = self.relative_position_index.view(-1) if self.rpe else None

        for i, blk in enumerate(self.blocks[-self.num_main_blocks:]):
            window_attn = self.window_size is not None and i not in self.global_indices
            Ws = self.window_size
            if window_attn:
                x, pad_hw = window_partition(x.reshape(B, Hp, Wp, -1), Ws)
                x = x.flatten(1, 2)
            x = checkpoint.checkpoint(blk, x, rpe_index) if self.use_checkpoint else blk(x, rpe_index)
            if window_attn:
                x = window_unpartition(x.reshape(x.size(0), Ws, Ws, -1), Ws, pad_hw, (Hp, Wp))
                x = x.flatten(1, 2)
            if i in self.out_indices:
                xp = x.permute(0, 2, 1).reshape(B, -1, Hp, Wp).contiguous()
                for _ in range(self.out_indices.count(i)):
                    features.append(xp)

        if self.with_fpn:
            ops = [self.fpn1, self.fpn2, self.fpn3, self.fpn4]
            for i in range(len(features)):
                features[i] = ops[i](features[i])

        if self.last_feat:
            return tuple(features), x
        return tuple(features)



import torch
import torch.nn.functional as F
from tqdm import tqdm
from loss.diceLoss import dice_coeff
from loss.bceLoss import BCE_loss


def eval_net(net, loader, device, n_val):
    """Evaluation without the densecrf with the dice coefficient"""
    net.eval()
    tot = 0

    with tqdm(total=n_val, desc='Validation round', unit='img', leave=False) as pbar:
        for batch in loader:
            imgs = batch['image']
            true_masks = batch['mask']

            imgs = imgs.to(device=device, dtype=torch.float32)
            mask_type = torch.float32 if net.n_classes == 1 else torch.long
            true_masks = true_masks.to(device=device, dtype=mask_type)
            mask_pred = net(imgs)

            for true_mask, pred in zip(true_masks, mask_pred):
                pred = (pred > 0.5).float()
                if net.n_classes > 1:
                    tot += F.cross_entropy(pred.unsqueeze(dim=0), true_mask.unsqueeze(dim=0)).item()
                else:
                    tot += dice_coeff(pred, true_mask.squeeze(dim=1)).item()
            
            pbar.update(imgs.shape[0])
    return tot / n_val



import cv2 
import glob 
import os
import os.path as osp 
# from ..converters.utils import xywh2xyxy
def xywh2xyxy(imgsz, xywh):

    if not len(imgsz) >= 2:
        raise RuntimeError(f"imgsz should be [height, width, channel] or [height, width]")
    elif imgsz[0] <= 3:
        raise RuntimeError(f"imgsz should be [height, width, channel] or [height, width]")

    if isinstance(xywh, list):
        x, y, w, h = xywh[0], xywh[1], xywh[2], xywh[3]
    else:
        raise RuntimeError(f"xywh must be list such as [x, y, w, h]")

    dh, dw = imgsz[0], imgsz[1]
    l = ((x - w / 2) * dw) # x0
    r = ((x + w / 2) * dw) # x1
    t = ((y - h / 2) * dh) # y0
    b = ((y + h / 2) * dh) # y1
    
    if l < 0:
        l = 0
    if r > dw - 1:
        r = dw - 1
    if t < 0:
        t = 0
    if b > dh - 1:
        b = dh - 1   

    return [l, t, r, b] # x0, y0, x1, y1


input_dir = '/HDD/datasets/projects/interojo/split_datasets_yolo'
output_dir = '/HDD/datasets/projects/interojo/vis'
image_ext = 'png'

if not osp.exists(output_dir):
    os.mkdir(output_dir)


folders = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "images/**")) if not osp.isfile(folder)]
for folder in folders:
    img_files = glob.glob(osp.join(input_dir, 'images', folder, f'*.{image_ext}'))
    
    _output_dir = osp.join(output_dir, folder)
    if not osp.exists(_output_dir):
        os.mkdir(_output_dir)
    
    for img_file in img_files:
        filename = osp.split(osp.splitext(img_file)[0])[-1]
        txt_file = osp.join(input_dir, 'labels', folder, filename + '.txt')
        assert osp.exists(txt_file), RuntimeError(f"There is no such label file for {txt_file}")

        image = cv2.imread(img_file)
        txt = open(txt_file, 'r')
        while True:
            line = txt.readline()
            if not line: break
            
            label_id, cx, cy, w, h = list(map(float, line.split(" ")))
            # print("xywh: ", cx, cy, w, h)
            xyxy = xywh2xyxy(image.shape, [cx, cy, w, h])
            # print("xyxy: ", xyxy)
            points, point = [], []
            for idx, __xyxy in enumerate(xyxy):
                if (idx)%2 == 0:
                    point = []
                    points.append(point)
                point.append(round(__xyxy, 1))

            x1, y1, x2, y2 = xyxy
            cv2.putText(image, str(label_id), (int(x1), int(y1)), cv2.FONT_HERSHEY_SIMPLEX, .5, (0, 0, 255), 1)
            cv2.rectangle(image, (int(x1), int(y1)), (int(x2), int(y2)), (0, 0, 255), 2)

        cv2.imwrite(osp.join(_output_dir, filename + '.png'), image)

            
        
        


########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################


import os
import pickle
import abc

import numpy as np
from torch.utils.data import Dataset


class DatasetBase(abc.ABC, Dataset):
    def __init__(self):
        self._camera = None
        self._default_preprocessor = lambda x: x
        self.preprocessor = self._default_preprocessor

    def filter_camera(self, camera):
        assert camera in self.cameras
        self._camera = camera
        return self

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self._camera = None

    @abc.abstractmethod
    def __len__(self):
        pass

    def __getitem__(self, idx):
        sample = {
            "image": self.load_image(idx),
            "label": self.load_label(idx),
            "name": self.load_name(idx),
        }

        if self.split != "train":
            # needed to compute mIoU on original image size
            sample["label_orig"] = sample["label"].copy()

        if self.with_input_orig:
            sample["image_orig"] = sample["image"].copy()
        sample = self.preprocessor(sample)

        return sample

    @property
    @abc.abstractmethod
    def cameras(self):
        pass

    @property
    @abc.abstractmethod
    def class_names(self):
        pass

    @property
    @abc.abstractmethod
    def class_names_without_void(self):
        pass

    @property
    @abc.abstractmethod
    def class_colors(self):
        pass

    @property
    @abc.abstractmethod
    def class_colors_without_void(self):
        pass

    @property
    @abc.abstractmethod
    def n_classes(self):
        pass

    @property
    @abc.abstractmethod
    def n_classes_without_void(self):
        pass

    @property
    @abc.abstractmethod
    def split(self):
        pass

    @property
    @abc.abstractmethod
    def source_path(self):
        pass

    @property
    @abc.abstractmethod
    def with_input_orig(self):
        pass

    @property
    def camera(self):
        return self._camera

    @abc.abstractmethod
    def load_image(self, idx):
        pass

    @abc.abstractmethod
    def load_label(self, idx):
        pass

    def color_label(self, label, with_void=True):
        if with_void:
            colors = self.class_colors
        else:
            colors = self.class_colors_without_void
        cmap = np.asarray(colors, dtype="uint8")

        return cmap[label]

    @staticmethod
    def static_color_label(label, colors):
        cmap = np.asarray(colors, dtype="uint8")
        return cmap[label]

    def compute_class_weights(self, weight_mode="median_frequency", c=1.02):
        assert weight_mode in ["median_frequency", "logarithmic", "linear"]

        # build filename
        class_weighting_filepath = os.path.join(
            self.source_path,
            f"weighting_{weight_mode}_" f"1+{self.n_classes_without_void}",
        )
        if weight_mode == "logarithmic":
            class_weighting_filepath += f"_c={c}"

        class_weighting_filepath += f"_{self.split}.pickle"

        if os.path.exists(class_weighting_filepath):
            class_weighting = pickle.load(open(class_weighting_filepath, "rb"))
            print(f"Using {class_weighting_filepath} as class weighting")
            return class_weighting

        print("Compute class weights")

        n_pixels_per_class = np.zeros(self.n_classes)
        n_image_pixels_with_class = np.zeros(self.n_classes)
        for i in range(len(self)):
            label = self.load_label(i)
            label[label == 19] = 16
            h, w = label.shape
            current_dist = np.bincount(label.flatten(), minlength=self.n_classes)
            n_pixels_per_class += current_dist

            # For median frequency we need the pixel sum of the images where
            # the specific class is present. (It only matters if the class is
            # present in the image and not how many pixels it occupies.)
            class_in_image = current_dist > 0
            n_image_pixels_with_class += class_in_image * h * w

            print(f"\r{i+1}/{len(self)}", end="")
        print()

        # remove void
        n_pixels_per_class = n_pixels_per_class[1:]
        n_image_pixels_with_class = n_image_pixels_with_class[1:]

        if weight_mode == "linear":
            class_weighting = n_pixels_per_class

        elif weight_mode == "median_frequency":
            frequency = n_pixels_per_class / n_image_pixels_with_class
            class_weighting = np.median(frequency) / frequency

        elif weight_mode == "logarithmic":
            probabilities = n_pixels_per_class / np.sum(n_pixels_per_class)
            class_weighting = 1 / np.log(c + probabilities)

        if np.isnan(np.sum(class_weighting)):
            print(f"n_pixels_per_class: {n_pixels_per_class}")
            print(f"n_image_pixels_with_class: {n_image_pixels_with_class}")
            print(f"class_weighting: {class_weighting}")
            raise ValueError("class weighting contains NaNs")

        with open(class_weighting_filepath, "wb") as f:
            pickle.dump(class_weighting, f)
        print(f"Saved class weights under {class_weighting_filepath}.")
        return class_weighting



from .base_runner import BaseRunner

            
class BaseTestRunner(BaseRunner):
    def test(self, cfgs_file=None, *args, **kwargs):
        
        self.set_configs(cfgs_file=cfgs_file, *args, **kwargs)
        self.set_variables()
        self.run()
    


import torch


def get_modules(use_v2):
    # We need a protected import to avoid the V2 warning in case just V1 is used
    if use_v2:
        import torchvision.transforms.v2
        import torchvision.tv_tensors
        import v2_extras

        return torchvision.transforms.v2, torchvision.tv_tensors, v2_extras
    else:
        import visionsuite.engines.utils.torch_utils.transforms as transforms

        return transforms, None, None


class TrainTransform:
    def __init__(
        self,
        resize,
        normalize,
        backend="pil",
        use_v2=False,
    ):
        T, tv_tensors, v2_extras = get_modules(use_v2)

        transforms = []
        backend = backend.lower()
        if backend == "tv_tensor":
            transforms.append(T.ToImage())
        elif backend == "tensor":
            transforms.append(T.PILToTensor())
        elif backend != "pil":
            raise ValueError(f"backend can be 'tv_tensor', 'tensor' or 'pil', but got {backend}")

        transforms += [T.Resize(size=(resize['height'], resize['width']))]

        if backend == "pil":
            transforms += [T.PILToTensor()]

        if use_v2:
            img_type = tv_tensors.Image if backend == "tv_tensor" else torch.Tensor
            transforms += [
                T.ToDtype(dtype={img_type: torch.float32, tv_tensors.Mask: torch.int64, "others": None}, scale=True)
            ]
        else:
            # No need to explicitly convert masks as they're magically int64 already
            transforms += [T.ToDtype(torch.float, scale=True)]

        if normalize == 'imagenet':
            transforms += [T.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))]
            
        if use_v2:
            transforms += [T.ToPureTensor()]

        self.transforms = T.Compose(transforms)

    def __call__(self, img, target):
        return self.transforms(img, target)


class ValTransform:
    def __init__(
        self, resize, normalize, backend="pil", use_v2=False
    ):
        T, _, _ = get_modules(use_v2)

        transforms = []
        backend = backend.lower()
        if backend == "tensor":
            transforms += [T.PILToTensor()]
        elif backend == "tv_tensor":
            transforms += [T.ToImage()]
        elif backend != "pil":
            raise ValueError(f"backend can be 'tv_tensor', 'tensor' or 'pil', but got {backend}")

        transforms += [T.Resize(size=(resize['height'], resize['width']))]

        if backend == "pil":
            # Note: we could just convert to pure tensors even in v2?
            transforms += [T.ToImage() if use_v2 else T.PILToTensor()]

        transforms += [T.ToDtype(torch.float, scale=True)]

        if normalize == 'imagenet':
            transforms += [T.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))]

        if use_v2:
            transforms += [T.ToPureTensor()]

        self.transforms = T.Compose(transforms)

    def __call__(self, img, target):
        return self.transforms(img, target)




from tensorflow import math
from tensorflow.keras.layers import Layer
import tensorflow.keras.backend as K


def gelu_(X):

    return 0.5*X*(1.0 + math.tanh(0.7978845608028654*(X + 0.044715*math.pow(X, 3))))

def snake_(X, beta):

    return X + (1/beta)*math.square(math.sin(beta*X))


class GELU(Layer):
    '''
    Gaussian Error Linear Unit (GELU), an alternative of ReLU
    
    Y = GELU()(X)
    
    ----------
    Hendrycks, D. and Gimpel, K., 2016. Gaussian error linear units (gelus). arXiv preprint arXiv:1606.08415.
    
    Usage: use it as a tf.keras.Layer
    
    
    '''
    def __init__(self, trainable=False, **kwargs):
        super(GELU, self).__init__(**kwargs)
        self.supports_masking = True
        self.trainable = trainable

    def build(self, input_shape):
        super(GELU, self).build(input_shape)

    def call(self, inputs, mask=None):
        return gelu_(inputs)

    def get_config(self):
        config = {'trainable': self.trainable}
        base_config = super(GELU, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))
    def compute_output_shape(self, input_shape):
        return input_shape

    
class Snake(Layer):
    '''
    Snake activation function $X + (1/b)*sin^2(b*X)$. Proposed to learn periodic targets.
    
    Y = Snake(beta=0.5, trainable=False)(X)
    
    ----------
    Ziyin, L., Hartwig, T. and Ueda, M., 2020. Neural networks fail to learn periodic functions 
    and how to fix it. arXiv preprint arXiv:2006.08195.
    
    '''
    def __init__(self, beta=0.5, trainable=False, **kwargs):
        super(Snake, self).__init__(**kwargs)
        self.supports_masking = True
        self.beta = beta
        self.trainable = trainable

    def build(self, input_shape):
        self.beta_factor = K.variable(self.beta, dtype=K.floatx(), name='beta_factor')
        if self.trainable:
            self._trainable_weights.append(self.beta_factor)

        super(Snake, self).build(input_shape)

    def call(self, inputs, mask=None):
        return snake_(inputs, self.beta_factor)

    def get_config(self):
        config = {'beta': self.get_weights()[0] if self.trainable else self.beta, 'trainable': self.trainable}
        base_config = super(Snake, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape
    


from visionsuite.engines.segmentation.utils.registry import LOSSES

def build_loss(**config):
    loss = LOSSES.get(config['type'], case_sensitive=config['case_sensitive'])(**config)

    return loss




#!/usr/bin/env python3

"""Custom EncoderDecoder that subclasses the original version.

- Support methods that depends on passing input image to the head.
- Need to use mmseg dataset class.
- Only supports mmseg registered LOSSES (SEG_LOSSES).
"""

import torch.nn as nn

from mmseg.core import add_prefix
from mmseg.models.segmentors import EncoderDecoder

from .. import builder
from ..builder import SEGMENTORS


@SEGMENTORS.register_module()
class CustomEncoderDecoder(EncoderDecoder):
    """Custom Encoder Decoder segmentors.

    - pass input image to decode head
    """

    def __init__(
        self,
        *args,
        pass_input_image=False,
        **kwargs,
    ):
        super(CustomEncoderDecoder, self).__init__(*args, **kwargs)

        # pass input image to decode head
        self.pass_input_image = pass_input_image

    def extract_feat(self, img):
        """Extract features from images."""
        x = self.backbone(img)
        if self.with_neck:
            x = self.neck(x)

        if self.pass_input_image:
            # HACK: appends image to the end of the inputs
            x = (*x, img)

        return x


@SEGMENTORS.register_module()
class SBCBEncoderDecoder(CustomEncoderDecoder):
    """SBCB Encoder Decoder Module.

    This module is only used to train the auxiliary model for SBCB.
    For inference, you can remove the auxiliary head using the config or
    load the weights into a normal EncoderDecoder model.
    """

    def __init__(self, *args, auxiliary_edge_head=None, **kwargs):
        super().__init__(*args, **kwargs)

        # initialize auxiliary edge head
        self._init_auxiliary_edge_head(auxiliary_edge_head)

    @property
    def with_auxiliary_edge_head(self):
        """bool: whether the segmentor has auxiliary edge head for SBCB"""
        return (
            hasattr(self, "auxiliary_edge_head")
            and self.auxiliary_edge_head is not None
        )

    def _init_auxiliary_edge_head(self, auxiliary_edge_head):
        if auxiliary_edge_head is not None:
            if isinstance(auxiliary_edge_head, list):
                self.auxiliary_edge_head = nn.ModuleList()
                for head_cfg in auxiliary_edge_head:
                    self.auxiliary_edge_head.append(builder.build_head(head_cfg))
            else:
                self.auxiliary_edge_head = builder.build_head(auxiliary_edge_head)

    def _auxiliary_edge_head_forward_train(self, x, img_metas, gt_semantic_edge):
        """Forward function for auxiliary edge head in training."""

        losses = dict()
        if isinstance(self.auxiliary_edge_head, nn.ModuleList):
            for idx, aux_head in enumerate(self.auxiliary_edge_head):
                loss_aux = aux_head.forward_train(
                    x, img_metas, gt_semantic_edge, self.train_cfg
                )
                losses.update(add_prefix(loss_aux, f"aux_{idx}"))
        else:
            loss_aux = self.auxiliary_edge_head.forward_train(
                x, img_metas, gt_semantic_edge, self.train_cfg
            )
            losses.update(add_prefix(loss_aux, "aux"))

        return losses

    def forward_train(
        self, img, img_metas, gt_semantic_seg, gt_semantic_edge=None, **kwargs
    ):
        x = self.extract_feat(img)

        losses = dict()

        loss_decode = self._decode_head_forward_train(
            x,
            img_metas,
            gt_semantic_seg,
        )
        losses.update(loss_decode)

        if self.with_auxiliary_head:
            loss_aux = self._auxiliary_head_forward_train(
                x,
                img_metas,
                gt_semantic_seg,
            )
            losses.update(loss_aux)

        if self.with_auxiliary_edge_head:
            assert (
                gt_semantic_edge is not None
            ), "auxiliary edge exists but gt_semantic_edge is None"
            loss_aux_edge = self._auxiliary_edge_head_forward_train(
                x,
                img_metas,
                gt_semantic_edge,
            )
            losses.update(loss_aux_edge)

        return losses



#!/usr/bin/env python3

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..builder import LOSSES


def weighted_binary_loss(
    edge,
    edge_label,
    alpha=1.0,
    beta=1.0,
    reduction="mean",
    ignore_index=255,
):
    # input edge dim=4 (b, 1, h, w)
    # input edge_label dim=4 (b, 1, h, w)
    pos_index = edge_label == 1
    neg_index = edge_label == 0
    ignore_index = edge_label > 1  # should be `ignore_index`

    weight = torch.Tensor(edge.size()).fill_(0)
    pos_num = pos_index.sum()
    neg_num = neg_index.sum()
    sum_num = pos_num + neg_num
    weight[pos_index] = alpha * neg_num / sum_num
    weight[neg_index] = beta * pos_num / sum_num
    weight[ignore_index] = 0
    weight = weight.to(edge.device)

    return F.binary_cross_entropy_with_logits(
        edge,
        edge_label.float(),
        weight,
        reduction=reduction,
    )


def balanced_binary_loss(
    edge,
    edge_label,
    reduction="mean",
    ignore_index=255,
    sensitivity=10,
):
    # input edge dim=4 (b, 1, h, w)
    # input edge_label dim=4 (b, 1, h, w)
    pos_index = edge_label == 1
    neg_index = edge_label == 0

    pos_num = pos_index.sum()
    neg_num = neg_index.sum()
    sum_num = pos_num + neg_num

    pos_weight = (neg_num / pos_num).clamp(min=1, max=sum_num) / sensitivity

    w = torch.tensor([pos_weight], device=edge.device)

    return F.binary_cross_entropy_with_logits(
        edge,
        edge_label.float(),
        reduction=reduction,
        pos_weight=w.reshape(1, 1, 1, 1),
    )


@LOSSES.register_module()
class BinaryEdgeLoss(nn.Module):
    def __init__(
        self,
        loss_weight=1.0,
        alpha=1.0,
        beta=1.0,
        loss_name="loss_binary_edge",
    ):
        super().__init__()
        self.loss_weight = loss_weight
        self._loss_name = loss_name
        self._alpha = alpha
        self._beta = beta

    def forward(
        self,
        edge,  # logits
        edge_label,
        weight=None,
        ignore_index=255,
        **kwargs,
    ):
        return self.loss_weight * weighted_binary_loss(
            edge=edge,
            edge_label=edge_label,
            alpha=self._alpha,
            beta=self._beta,
            reduction="mean",
            ignore_index=ignore_index,
        )

    @property
    def loss_name(self):
        return self._loss_name


@LOSSES.register_module()
class ConsensusBinaryEdgeLoss(nn.Module):
    def __init__(
        self,
        loss_weight=1.0,
        loss_name="loss_conbin_edge",
    ):
        super().__init__()
        self.loss_weight = loss_weight
        self._loss_name = loss_name

    def forward(
        self,
        edge,  # logits
        edge_label,
        weight=None,
        ignore_index=255,
        **kwargs,
    ):
        return self.loss_weight * weighted_binary_loss(
            edge=edge,
            edge_label=edge_label,
            alpha=1.0,
            beta=1.1,
            reduction="mean",
            ignore_index=ignore_index,
        )

    @property
    def loss_name(self):
        return self._loss_name




from visionsuite.cores.roboflow.configs.train_config import TrainConfig as YoloV8TrainConfig
from visionsuite.cores.roboflow.configs.train_config_rtdetr import TrainConfig as RtdetrTrainConfig

class Agent:
    
    @classmethod
    def get_tasks(cls):
        return ['classification', 'hbb detection', 'obb detection', 'instance segmentation']
    
    @classmethod
    def get_models_by_task(cls, task):
        if task == 'clssification':
            NotImplementedError 
        elif task == 'hbb detection':
            return cls.get_hbb_det_models()
        elif task == 'obb detection':
            return cls.get_obb_det_models()
        elif task == 'instance segmentation':
            return cls.get_seg_models()
        
    @classmethod
    def get_cls_models(cls):
        return ['yolov8']
    
    @classmethod
    def get_hbb_det_models(cls):
        return ['yolov8', 'yolov10', 'rtdetr']
    
    @classmethod
    def get_obb_det_models(cls):
        return ['yolov8', 'yolov10']
    
    @classmethod
    def get_seg_models(cls):
        return ['yolov8']
    
    @staticmethod
    def get_schema_by_model(model):
        if model == 'yolov8':
            return {"TrainConfig": YoloV8TrainConfig.model_json_schema(mode='serialization')}
        elif model == 'yolov10':
            return {"TrainConfig": YoloV8TrainConfig.model_json_schema(mode='serialization')}
        elif model == 'rtdetr':
            return {"TrainConfig": RtdetrTrainConfig.model_json_schema(mode='serialization')}
        else:
            return {'TrainConfig': {}}



import os.path as osp 
import glob 
import os
import json
from shutil import copyfile
from tqdm import tqdm


def get_classes_info(input_dir):
    modes = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "**")) if not osp.isfile(folder)]

    if len(modes) == 0:
        modes = ['./']
        
    classes = {}
        
    for mode in modes:
        json_files = glob.glob(osp.join(input_dir, mode, '*.json'))
        
        classes[mode] = {}
        for json_file in tqdm(json_files, desc=mode):
            with open(json_file, 'r') as jf:
                anns = json.load(jf)['shapes']
                
            for ann in anns:
                label = ann['label']

                if label not in classes[mode]:
                    classes[mode][label] = 1
                else:
                    classes[mode][label] += 1    
                
    return classes 

def balance_dataset(input_dir, output_dir, counts, image_format):
    modes = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "**")) if not osp.isfile(folder)]

    if len(modes) == 0:
        modes = ['./']
        
    if not osp.exists(output_dir):
        os.mkdir(output_dir)    
    
    for mode in modes:
        json_files_to_move = []
        json_files = glob.glob(osp.join(input_dir, mode, '*.json'))
        
        _output_dir = osp.join(output_dir, mode)
        if not osp.exists(_output_dir):
            os.mkdir(_output_dir)    

        for json_file in tqdm(json_files, desc=mode):
            with open(json_file, 'r') as jf:
                anns = json.load(jf)['shapes']
                
            for ann in anns:
                label = ann['label']
                
                if counts[label] > 0:
                    counts[label] -= 1
                    
                    json_files_to_move.append(json_file)
                    
        for json_file_to_move in set(json_files_to_move):
            filename = osp.split(osp.splitext(json_file_to_move)[0])[-1]
            img_file = osp.splitext(json_file_to_move)[0] + f".{image_format}"
            copyfile(img_file, osp.join(_output_dir, filename + f".{image_format}"))
            copyfile(json_file_to_move, osp.join(_output_dir, filename + ".json"))             
    
    
if __name__ == '__main__':
    input_dir = '/HDD/datasets/projects/Tenneco/Metalbearing/inner/split_dataset'
    output_dir = '/HDD/datasets/projects/Tenneco/Metalbearing/inner/split_dataset_balanced'
    
    classes_info = get_classes_info(input_dir)
    
    print(classes_info)
    
    counts = {'STABBED': 100, 'STABBED_C': 100, 'SCRATCH': 100, 'STABBED_P': 55}   
    image_format = 'bmp' 

    balance_dataset(input_dir, output_dir, counts, image_format)
    
    
    classes_info = get_classes_info(output_dir)
    print(classes_info)



import glob 
import os
import os.path as osp 
import json 
import numpy as np
from visionsuite.utils.dataset.converters.utils import xyxy2xywh
from tqdm import tqdm
import warnings


def labelme2yolo_hbb(input_dir, output_dir, copy_image=True, 
                     image_ext='bmp', image_width=None, image_height=None):

    if not osp.exists(output_dir):
        os.mkdir(output_dir)
        
    folders = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "**")) if not osp.isfile(folder)]

    class2idx = {}
    for folder in folders:
        _output_labels_dir = osp.join(output_dir, 'labels', folder)
        if not osp.exists(_output_labels_dir):
            os.makedirs(_output_labels_dir)

        json_files = glob.glob(osp.join(input_dir, folder, '*.json'))
        print(f"There are {len(json_files)} json files")

        for json_file in tqdm(json_files, desc=folder):
            filename = osp.split(osp.splitext(json_file)[0])[-1]
            txt = open(osp.join(_output_labels_dir, filename + '.txt'), 'w')
            with open(json_file, 'r') as jf:
                _anns = json.load(jf)
                anns = _anns['shapes']
                
                if 'imageHeight' in _anns and 'imageWidth' in _anns:
                    image_height, image_width = _anns['imageHeight'], _anns['imageWidth']
                
            if copy_image:
                import cv2
                from shutil import copyfile
                img_file = osp.join(input_dir, folder, filename + f'.{image_ext}')
                
                assert osp.exists(img_file), ValueError(f"There is no such image: {img_file}")
                
                image = cv2.imread(img_file)
                image_width = image.shape[1]
                image_height = image.shape[0]
                _output_image_dir = osp.join(output_dir, 'images', folder)
                if not osp.exists(_output_image_dir):
                    os.makedirs(_output_image_dir)
                copyfile(img_file, osp.join(_output_image_dir, filename + f'.{image_ext}'))
                
            if len(anns) != 0:
                for ann in anns:
                    shape_type = ann['shape_type']
                    label = ann['label']
                    if label not in class2idx.keys():
                        class2idx.update({label: len(class2idx)})
                    points = ann['points']
                    xyxy = []
                    
                    if shape_type == 'point':
                        warnings.warn(f"Positive point is included")
                    
                    elif shape_type == 'rectangle':
                        for point in points:
                            xyxy.append(point[0])
                            xyxy.append(point[1])

                        if len(xyxy) > 4:
                            raise RuntimeError(f"shape type is rectangle, but there are more than 4 points")

                    elif shape_type == 'polygon' or shape_type == 'Watershed':
                        xs, ys = [], []
                        for point in points:
                            xs.append(point[0])
                            ys.append(point[1])

                        xyxy.append(np.max(xs))
                        xyxy.append(np.max(ys))
                        xyxy.append(np.min(xs))
                        xyxy.append(np.min(ys))
                    else: 
                        print(f"NotImplemented shape: {shape_type} for {json_file}")
                        continue

                        
                    assert image_width is not None and image_height is not None, RuntimeError(f"Image width is {image_width} and image height is {image_height}")
                    xywh = xyxy2xywh([image_height, image_width], xyxy)
                    txt.write(str(class2idx[label]) + ' ')
                    for kdx in range(len(xywh)):
                        if kdx == len(xywh) -1:
                            txt.write(str(round(xywh[kdx], 3)))
                        else:
                            txt.write(str(round(xywh[kdx], 3)) + ' ')
                    txt.write('\n')
            
        txt.close()

    txt = open(osp.join(output_dir, 'classe2idx.txt'), 'w')
    for key, val in class2idx.items():
        txt.write(f'{key}: {val}\n')
    txt.close()
    
    txt = open(osp.join(output_dir, 'idx2class.txt'), 'w')
    for key, val in class2idx.items():
        txt.write(f'{val}: {key}\n')
    txt.close()
            
            
def labelme2yolo_iseg(input_dir, output_dir, image_ext,
                                       copy_image=True, image_width=None, image_height=None):

    if not osp.exists(output_dir):
        os.mkdir(output_dir)
        
    folders = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "**")) if not osp.isfile(folder)]


    class2idx = {}
    for folder in folders:
        _output_labels_dir = osp.join(output_dir, 'labels', folder)
        if not osp.exists(_output_labels_dir):
            os.makedirs(_output_labels_dir)

        json_files = glob.glob(osp.join(input_dir, folder, '*.json'))
        print(f"There are {len(json_files)} json files")

        for json_file in tqdm(json_files, desc=folder):
            filename = osp.split(osp.splitext(json_file)[0])[-1]
            txt = open(osp.join(_output_labels_dir, filename + '.txt'), 'w')
            with open(json_file, 'r') as jf:
                anns = json.load(jf)['shapes']
                
            if copy_image:
                import cv2
                from shutil import copyfile
                img_file = osp.join(input_dir, folder, filename + f'.{image_ext}')
                assert osp.exists(img_file), ValueError(f"There is no such image: {img_file}")
                
                image = cv2.imread(img_file)
                image_width = image.shape[1]
                image_height = image.shape[0]
                _output_image_dir = osp.join(output_dir, 'images', folder)
                if not osp.exists(_output_image_dir):
                    os.makedirs(_output_image_dir)
                copyfile(img_file, osp.join(_output_image_dir, filename + f'.{image_ext}'))
                
                
            assert image_width != None and image_height != None, ValueError(f"The input size (width, height) must be assigned")
                
            if len(anns) != 0:
                for ann in anns:
                    shape_type = ann['shape_type']
                    label = ann['label']
                    if label not in class2idx.keys():
                        class2idx.update({label: len(class2idx)})
                    points = ann['points']
                    if shape_type == 'point' or len(points) <= 2:
                        continue
                    assert len(points) >= 3, RuntimeError(f"The number of polygon points must be more than 3, not {len(points)} with {shape_type}")

                    txt.write(str(class2idx[label]) + ' ')
                    for idx, point in enumerate(points):
                        if idx == len(points) -1:
                            txt.write(f'{round(point[0]/image_width, 3)} {round(point[1]/image_height, 3)}')
                        else:
                            txt.write(f'{round(point[0]/image_width, 3)} {round(point[1]/image_height, 3)} ')
                    txt.write('\n')
            
        txt.close()

    txt = open(osp.join(output_dir, 'classes2idx.txt'), 'w')
    for key, val in class2idx.items():
        txt.write(f'{val}: {key}\n')
    txt.close()
    
    txt = open(osp.join(output_dir, 'idx2class.txt'), 'w')
    for key, val in class2idx.items():
        txt.write(f'{val}: {key}\n')
    txt.close()
    
def labelme2yolo_obb(input_dir: str, output_dir: str=None, copy_image=True, image_ext='bmp'):
    from visionsuite.utils.dataset.converters.labelme2dota import convert_labelme2dota
    from visionsuite.utils.dataset.converters.dota2yolo import convert_dota2yolo_obb

    output_dir = input_dir + '_dota'

    convert_labelme2dota(input_dir, output_dir, copy_image=copy_image, image_ext=image_ext)
    print("** FINISED to convert labelme2dota")

    input_dir = output_dir
    output_dir = input_dir + '_yolo_obb'

    convert_dota2yolo_obb(input_dir, output_dir, copy_image=copy_image, image_ext=image_ext)
    print("** FINISED to convert dota2yolo_obb")


# Copyright (c) Open-MMLab. All rights reserved.
from .checkpoint import save_checkpoint
from .epoch_based_runner import EpochBasedRunnerAmp


__all__ = [
    'EpochBasedRunnerAmp', 'save_checkpoint'
]



def check_label_roi(roi, label):

    for _label in label:
        points = _label["points"]
        for point in points:
            print(points, roi)
            assert roi[0] <= point[0] and point[0] <= roi[2], ValueError(
                f"Point ({point}) is out of roi ({roi})"
            )
            assert roi[1] <= point[1] and point[1] <= roi[3], ValueError(
                f"Point ({point}) is out of roi ({roi})"
            )



import os.path as osp
import requests

def download_weights_from_url(url, output_filename=None):
    
    try:
        print(f"Now downloading {url}")
        response = requests.get(url)
        
        if not osp.exists(output_filename):
            with open(output_filename, 'wb') as file:
                file.write(response.content)
            print(f"File downloaded and saved as {output_filename}")
        else:
            print(f"Already exists at {output_filename}")
            
        return True
        
    except Exception as error:
        raise RuntimeError(f"[ERROR] There is no such url: {url} - error")
    
if __name__ == '__main__':
    from ultralytics import YOLO
    url = "https://huggingface.co/spaces/hamhanry/YOLOv10-OBB/resolve/main/pretrained/yolov10s-obb.pt"
    output_filename = "/HDD/etc/yolov10s-obb.pt"
    download_weights_from_url(url, output_filename)
    
    
    # https://huggingface.co/spaces/hamhanry/YOLOv10-OBB/blob/main/pretrained/yolov10x-obb.pt



from __future__ import absolute_import

from keras_unet_collection.layer_utils import *
from keras_unet_collection.activations import GELU, Snake

from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model

def ResUNET_a_block(X, channel, kernel_size=3, dilation_num=1.0, activation='ReLU', batch_norm=False, name='res_a_block'):
    '''
    The "ResUNET-a" block
    
    ResUNET_a_block(X, channel, kernel_size=3, dilation_num=1.0, activation='ReLU', batch_norm=False, name='res_a_block')
    
    ----------
    Diakogiannis, F.I., Waldner, F., Caccetta, P. and Wu, C., 2020. Resunet-a: a deep learning framework for 
    semantic segmentation of remotely sensed data. ISPRS Journal of Photogrammetry and Remote Sensing, 162, pp.94-114.
    
    Input
    ----------
        X: input tensor.
        channel: number of convolution filters.
        kernel_size: size of 2-d convolution kernels.
        dilation_num: an iterable that defines dilation rates of convolutional layers.
                      stacks of conv2d is expected as `len(dilation_num)`.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
    
    '''
    
    X_res = []
    
    for i, d in enumerate(dilation_num):
        
        X_res.append(CONV_stack(X, channel, kernel_size=kernel_size, stack_num=2, dilation_rate=d, 
                                activation=activation, batch_norm=batch_norm, name='{}_stack{}'.format(name, i)))
        
    if len(X_res) > 1:
        return add(X_res)
    
    else:
        return X_res[0]


def ResUNET_a_right(X, X_list, channel, kernel_size=3, dilation_num=[1,], 
                    activation='ReLU', unpool=True, batch_norm=False, name='right0'):
    '''
    The decoder block of ResUNet-a
    
    ResUNET_a_right(X, X_list, channel, kernel_size=3, dilation_num=[1,], 
                    activation='ReLU', unpool=True, batch_norm=False, name='right0')
    
    Input
    ----------
        X: input tensor.
        X_list: a list of other tensors that connected to the input tensor.
        channel: number of convolution filters.
        kernel_size: size of 2-d convolution kernels.
        dilation_num: an iterable that defines dilation rates of convolutional layers.
                      stacks of conv2d is expected as `len(dilation_num)`.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.
        batch_norm: True for batch normalization, False otherwise.
        name: name of the created keras layers.
        
    Output
    ----------
        X: output tensor.

    
    '''
    
    pool_size = 2
    
    X = decode_layer(X, channel, pool_size, unpool, 
                     activation=activation, batch_norm=batch_norm, name='{}_decode'.format(name))
        
    # <--- *stacked convolutional can be applied here
    X = concatenate([X,]+X_list, axis=3, name=name+'_concat')
    
    # Stacked convolutions after concatenation 
    X = ResUNET_a_block(X, channel, kernel_size=kernel_size, dilation_num=dilation_num, activation=activation, 
                        batch_norm=batch_norm, name='{}_resblock'.format(name))
     
    return X

def resunet_a_2d_base(input_tensor, filter_num, dilation_num,
                      aspp_num_down=256, aspp_num_up=128, activation='ReLU',
                      batch_norm=True, pool=True, unpool=True, name='resunet'):
    '''
    The base of ResUNet-a
    
    resunet_a_2d_base(input_tensor, filter_num, dilation_num,
                      aspp_num_down=256, aspp_num_up=128, activation='ReLU',
                      batch_norm=True, pool=True, unpool=True, name='resunet')
                          
    ----------
    Diakogiannis, F.I., Waldner, F., Caccetta, P. and Wu, C., 2020. Resunet-a: a deep learning framework for 
    semantic segmentation of remotely sensed data. ISPRS Journal of Photogrammetry and Remote Sensing, 162, pp.94-114.
    
    Input
    ----------
        input_tensor: the input tensor of the base, e.g., `keras.layers.Inpyt((None, None, 3))`.
        filter_num: a list that defines the number of filters for each \
                    down- and upsampling levels. e.g., `[64, 128, 256, 512]`.
                    The depth is expected as `len(filter_num)`.
        dilation_num: an iterable that defines the dilation rates of convolutional layers.
                      Diakogiannis et al. (2020) suggested `[1, 3, 15, 31]`.
                      * This base function requires `len(filter_num) == len(dilation_num)`.
                      Explicitly defining dilation rates for each down-/upsampling level.
        aspp_num_down: number of Atrous Spatial Pyramid Pooling (ASPP) layer filters after the last downsampling block.
        aspp_num_up: number of ASPP layer filters after the last upsampling block.                 
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        batch_norm: True for batch normalization.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.                   
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
        
    * Downsampling is achieved through strided convolutional layers with 1-by-1 kernels in Diakogiannis et al., (2020), 
      and is here is achieved either with pooling layers or strided convolutional layers with 2-by-2 kernels.
    * If this base function is involved in network training, then the input shape cannot have NoneType.
    * `dilation_num` should be provided as 2d iterables, with the second dimension matches the model depth.
      e.g., for `len(filter_num) = 4`, dilation_num can be provided as: `[[1, 3, 15, 31], [1, 3, 15], [1,], [1,]]`.
      
    '''
    
    pool_size = 2
    
    activation_func = eval(activation)
    
    depth_ = len(filter_num)
    X_skip = []
    
    # ----- #
    # rejecting auto-mode from this base function
    if isinstance(dilation_num[0], int):
        raise ValueError('`resunet_a_2d_base` does not support automated determination of `dilation_num`.')
    else:
        dilation_ = dilation_num
    # ----- #
    
    X = input_tensor
    
    # input mapping with 1-by-1 conv
    X = Conv2D(filter_num[0], 1, 1, dilation_rate=1, padding='same', 
               use_bias=True, name='{}_input_mapping'.format(name))(X)
    X = activation_func(name='{}_input_activation'.format(name))(X)
    X_skip.append(X)
    # ----- #
    
    X = ResUNET_a_block(X, filter_num[0], kernel_size=3, dilation_num=dilation_[0], 
                        activation=activation, batch_norm=batch_norm, name='{}_res0'.format(name)) 
    X_skip.append(X)

    for i, f in enumerate(filter_num[1:]):
        ind_ = i+1
        
        X = encode_layer(X, f, pool_size, pool, activation=activation, 
                         batch_norm=batch_norm, name='{}_down{}'.format(name, i))
        X = ResUNET_a_block(X, f, kernel_size=3, dilation_num=dilation_[ind_], activation=activation, 
                            batch_norm=batch_norm, name='{}_resblock_{}'.format(name, ind_))
        X_skip.append(X)

    X = ASPP_conv(X, aspp_num_down, activation=activation, batch_norm=batch_norm, name='{}_aspp_bottom'.format(name))

    X_skip = X_skip[:-1][::-1]
    dilation_ = dilation_[:-1][::-1]
    
    for i, f in enumerate(filter_num[:-1][::-1]):

        X = ResUNET_a_right(X, [X_skip[i],], f, kernel_size=3, activation=activation, dilation_num=dilation_[i], 
                            unpool=unpool, batch_norm=batch_norm, name='{}_up{}'.format(name, i))

    X = concatenate([X_skip[-1], X], name='{}_concat_out'.format(name))

    X = ASPP_conv(X, aspp_num_up, activation=activation, batch_norm=batch_norm, name='{}_aspp_out'.format(name))
    
    return X


def resunet_a_2d(input_size, filter_num, dilation_num, n_labels,
                 aspp_num_down=256, aspp_num_up=128, activation='ReLU', output_activation='Softmax', 
                 batch_norm=True, pool=True, unpool=True, name='resunet'):
    '''
    ResUNet-a
    
    resunet_a_2d(input_size, filter_num, dilation_num, n_labels,
                 aspp_num_down=256, aspp_num_up=128, activation='ReLU', output_activation='Softmax', 
                 batch_norm=True, pool=True, unpool=True, name='resunet')
                 
    ----------
    Diakogiannis, F.I., Waldner, F., Caccetta, P. and Wu, C., 2020. Resunet-a: a deep learning framework for 
    semantic segmentation of remotely sensed data. ISPRS Journal of Photogrammetry and Remote Sensing, 162, pp.94-114.
    
    Input
    ----------
        input_size: the size/shape of network input, e.g., `(128, 128, 3)`.
        filter_num: a list that defines the number of filters for each \
                    down- and upsampling levels. e.g., `[64, 128, 256, 512]`.
                    The depth is expected as `len(filter_num)`.
        dilation_num: an iterable that defines the dilation rates of convolutional layers.
                      Diakogiannis et al. (2020) suggested `[1, 3, 15, 31]`.
                      * `dilation_num` can be provided as 2d iterables, with the second dimension matches 
                      the model depth. e.g., for len(filter_num) = 4; dilation_num can be provided as: 
                      `[[1, 3, 15, 31], [1, 3, 15], [1,], [1,]]`.
                      * If `dilation_num` is not provided per down-/upsampling level, then the automated
                      determinations will be applied.
        n_labels: number of output labels.
        aspp_num_down: number of Atrous Spatial Pyramid Pooling (ASPP) layer filters after the last downsampling block.
        aspp_num_up: number of ASPP layer filters after the last upsampling block.  
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        output_activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interface or 'Sigmoid'.
                           Default option is 'Softmax'.
                           if None is received, then linear activation is applied.
        batch_norm: True for batch normalization.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.                  
        name: prefix of the created keras layers.
        
    Output
    ----------
        model: a keras model.
        
    * Downsampling is achieved through strided convolutional layers with 1-by-1 kernels in Diakogiannis et al., (2020), 
      and is here is achieved either with pooling layers or strided convolutional layers with 2-by-2 kernels.
    * `resunet_a_2d` does not support NoneType input shape.
    
    '''
    
    activation_func = eval(activation)
    depth_ = len(filter_num)
    
    X_skip = []
    
    # input_size cannot have None
    if input_size[0] is None or input_size[1] is None:
        raise ValueError('`resunet_a_2d` does not support NoneType input shape')
        
    # ----- #
    if isinstance(dilation_num[0], int):
        print("Received dilation rates: {}".format(dilation_num))
    
        deep_ = (depth_-2)//2
        dilation_ = [[] for _ in range(depth_)]
        
        print("Received dilation rates are not defined on a per downsampling level basis.")
        print("Automated determinations are applied with the following details:")
        
        for i in range(depth_):
            if i <= 1:
                dilation_[i] += dilation_num
            elif i > 1 and i <= deep_+1:
                dilation_[i] += dilation_num[:-1]
            else:
                dilation_[i] += [1,]
            print('\tdepth-{}, dilation_rate = {}'.format(i, dilation_[i]))
            
    else:
        dilation_ = dilation_num
    # ----- #
    
    IN = Input(input_size)
    
    # base
    X = resunet_a_2d_base(IN, filter_num, dilation_,
                          aspp_num_down=aspp_num_down, aspp_num_up=aspp_num_up, activation=activation,
                          batch_norm=batch_norm, pool=pool, unpool=unpool, name=name)
    
    OUT = CONV_output(X, n_labels, kernel_size=1, activation=output_activation, name='{}_output'.format(name))

    model = Model([IN], [OUT,], name='{}_model'.format(name))
    
    return model




import inspect
from typing import List

import torch
from visionsuite.engines.utils.registry import LOSSES, Registry


def register_torch_losses() -> List[str]:
    TORCH_LOSSES = Registry("torch_loss", parent=LOSSES, scope="visionsuite.engines.utils.registry.losses")
    for module_name in dir(torch.nn):
        if module_name.startswith("__"):
            continue
        
        if module_name.endswith("Loss"):
            _object = getattr(torch.nn, module_name)
        else:
            continue
        if inspect.isclass(_object):  # Check if it's a class
            # Optionally check if it's a subclass of nn.Module or specific loss classes
            if issubclass(_object, torch.nn.Module): 
                TORCH_LOSSES.register(module=_object)

    return TORCH_LOSSES


TORCH_LOSSES = register_torch_losses()




import inspect
from typing import List

import torch
from visionsuite.engines.utils.registry import OPTIMIZERS, Registry


def register_torch_optimizers() -> List[str]:
    TORCH_OPTIMIZERS = Registry("torch_optimizer", parent=OPTIMIZERS, scope="visionsuite.engines.utils.registry.optims")
    for module_name in dir(torch.optim):
        if module_name.startswith("__"):
            continue
        
        _object = getattr(torch.optim, module_name)
        if inspect.isclass(_object) and issubclass(_object, torch.optim.Optimizer):
            TORCH_OPTIMIZERS.register(module=_object)

    return TORCH_OPTIMIZERS


TORCH_OPTIMIZERS = register_torch_optimizers()



from collections import deque 
import torch 
from visionsuite.engines.utils.torch_utils.dist import reduce_across_processes

class SmoothedValue:
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{median:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    def synchronize_between_processes(self):
        """
        Warning: does not synchronize the deque!
        """
        t = reduce_across_processes([self.count, self.total])
        t = t.tolist()
        self.count = int(t[0])
        self.total = t[1]

    @property
    def median(self):
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        return self.fmt.format(
            median=self.median, avg=self.avg, global_avg=self.global_avg, max=self.max, value=self.value
        )



import os
import sys
import logging
import torch
import numpy as np

from os.path import splitext
from os import listdir
from glob import glob
from torch.utils.data import Dataset
from PIL import Image
import os.path as osp


class BasicDataset(Dataset):
    def __init__(self, unet_type, imgs_dir, masks_dir, scale=1):
        self.unet_type = unet_type
        self.imgs_dir = imgs_dir
        self.masks_dir = masks_dir
        self.scale = scale
        
        assert 0 < scale <= 1, 'Scale must be between 0 and 1'
        self.img_files = glob(osp.join(imgs_dir, "*.bmp"))
        logging.info(f'Creating dataset with {len(self.img_files)} examples')


    def __len__(self):
        return len(self.img_files)


    @classmethod
    def preprocess(cls, unet_type, pil_img, scale):
        w, h = pil_img.size
        newW, newH = int(scale * w), int(scale * h)
        assert newW > 0 and newH > 0, 'Scale is too small'

        if unet_type != 'v3':
            pil_img = pil_img.resize((newW, newH))
        else:
            new_size = int(scale * 640)
            pil_img = pil_img.resize((new_size, new_size))

        img_nd = np.array(pil_img)
        if len(img_nd.shape) == 2:
            img_nd = np.expand_dims(img_nd, axis=2)

        # HWC to CHW
        img_trans = img_nd.transpose((2, 0, 1))
        if img_trans.max() > 1:
            img_trans = img_trans / 255
        return img_trans


    def __getitem__(self, i):
        img_file = self.img_files[i]
        mask_file = osp.join(self.masks_dir, osp.split(osp.splitext(img_file)[0])[-1] + '.bmp')

        assert osp.exists(img_file), ValueError(f'THere is no such image: {img_file}')
        assert osp.exists(mask_file), ValueError(f'THere is no such mask: {mask_file}')

        mask = Image.open(mask_file)
        img = Image.open(img_file)


        img = self.preprocess(self.unet_type, img, self.scale)
        mask = self.preprocess(self.unet_type, mask, self.scale)
        
        
        return {'image': torch.from_numpy(img), 'mask': torch.from_numpy(mask)}



#!/usr/bin/env python3

# SBCB auxiliaries
from .casenet_head import GeneralizedCASENetHead, AuxCASENetHead
from .dff_head import GeneralizedDFFHead, AuxDFFHead
from .dds_head import GeneralizedDDSHead, AuxDDSHead

# Deocde heads
from .mod_sep_aspp_head import ModDepthwiseSeparableASPPHead


__all__ = [
    "GeneralizedCASENetHead",
    "AuxCASENetHead",
    "GeneralizedDFFHead",
    "AuxDFFHead",
    "GeneralizedDDSHead",
    "AuxDDSHead",
    "ModDepthwiseSeparableASPPHead",
]



import argparse
import logging
import os
import sys
import math
import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torch.optim.lr_scheduler as lr_scheduler
from torch import optim
from torch.utils.tensorboard import SummaryWriter
from torch.utils.data import DataLoader, random_split
# from apex import amp
from tqdm import tqdm
from unet import UNet
from unet import UNet2Plus
from unet import UNet3Plus, UNet3Plus_DeepSup, UNet3Plus_DeepSup_CGM
from utils.dataset import BasicDataset
from utils.eval import eval_net

os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'

dir_img = '/HDD/_projects/benchmark/semantic_segmentation/new_model/datasets/patches'
dir_mask = '/HDD/_projects/benchmark/semantic_segmentation/new_model/datasets/masks'
dir_checkpoint = '/HDD/_projects/benchmark/semantic_segmentation/new_model/outputs/unet3p_torch'

def custom_collate_fn(batch):
    # 데이터를 float32로 변환
    batch = [{k: v.float() if torch.is_tensor(v) and v.dtype == torch.uint8 else v for k, v in d.items()} for d in batch]
    return torch.utils.data.dataloader.default_collate(batch)

def train_net(unet_type, net, device, epochs=5, batch_size=1, lr=0.1, val_percent=0.1, save_cp=True, img_scale=0.5):
    dataset = BasicDataset(unet_type, dir_img, dir_mask, img_scale)
    n_val = int(len(dataset) * val_percent)
    n_train = len(dataset) - n_val

    train, val = random_split(dataset, [n_train, n_val])
    
    train_loader = DataLoader(train, batch_size=batch_size, shuffle=True, num_workers=8, pin_memory=True,
                              collate_fn=custom_collate_fn)
    val_loader = DataLoader(val, batch_size=batch_size, shuffle=False, num_workers=8, pin_memory=True,
                            collate_fn=custom_collate_fn)
    
    writer = SummaryWriter(comment=f'LR_{lr}_BS_{batch_size}_SCALE_{img_scale}')
    global_step = 0

    logging.info(f'''Starting training:
                     UNet type:       {unet_type}
                     Epochs:          {epochs}
                     Batch size:      {batch_size}
                     Learning rate:   {lr}
                     Dataset size:    {len(dataset)}
                     Training size:   {n_train}
                     Validation size: {n_val}
                     Checkpoints:     {save_cp}
                     Device:          {device.type}
                     Images scaling:  {img_scale}''')

    optimizer = optim.RMSprop(net.parameters(), lr=lr, weight_decay=1e-8)
    # model, optimizer = amp.initialize(net, optimizer, opt_level="O1")
    model, optimizer = net, optimizer

    # Scheduler https://arxiv.org/pdf/1812.01187.pdf
    lf = lambda x: (((1 + math.cos(x * math.pi / epochs)) / 2) ** 1.0) * 0.95 + 0.05 #cosine
    scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lf)
    scheduler.last_epoch = global_step

    if net.n_classes > 1:
        criterion = nn.CrossEntropyLoss()
    else:
        criterion = nn.BCEWithLogitsLoss()
    
    lrs = []
    best_loss = 10000
    for epoch in range(epochs):
        cur_lr = optimizer.param_groups[0]['lr']
        print('\nEpoch=', (epoch + 1), ' lr=', cur_lr)
        net.train()
        epoch_loss = 0

        with tqdm(total=n_train, desc=f'Epoch {epoch + 1}/{epochs}', unit='img') as pbar:
            for batch in train_loader:
                imgs = batch['image']
                true_masks = batch['mask']
                
                assert imgs.shape[1] == net.n_channels, f'Network has been defined with {net.n_channels} input channels, ' \
                    f'but loaded images have {imgs.shape[1]} channels. Please check that the images are loaded correctly.'

                imgs = imgs.to(device=device, dtype=torch.float32)
                mask_type = torch.float32 if net.n_classes == 1 else torch.long
                true_masks = true_masks.to(device=device, dtype=mask_type)

                masks_pred = net(imgs)
                loss = criterion(masks_pred, true_masks)
                epoch_loss += loss.item()
                writer.add_scalar('Loss/train', loss.item(), global_step)
                pbar.set_postfix(**{'loss (batch)': loss.item()})

                optimizer.zero_grad()
                loss.backward()
                # with amp.scale_loss(loss, optimizer) as scaled_loss:
                #     scaled_loss.backward()
                # scaled_loss.backward()
                optimizer.step()

                pbar.update(imgs.shape[0])
                global_step += 1
                dataset_len = len(dataset)
                
                a1 = dataset_len // 10 if dataset_len // 10 > 0 else 1
                a2 = dataset_len / 10 if dataset_len / 10 > 0 else 1
                b1 = global_step % a1
                b2 = global_step % a2

                if global_step % (len(dataset) // (10 * batch_size)) == 0:
                    val_score = eval_net(net, val_loader, device, n_val)
                    if net.n_classes > 1:
                        logging.info('Validation cross entropy: {}'.format(val_score))
                        writer.add_scalar('Loss/test', val_score, global_step)
                    else:
                        logging.info('Validation Dice Coeff: {}'.format(val_score))
                        writer.add_scalar('Dice/test', val_score, global_step)

                    writer.add_images('images', imgs, global_step)
                    if net.n_classes == 1:
                        writer.add_images('masks/true', true_masks, global_step)
                        writer.add_images('masks/pred', torch.sigmoid(masks_pred) > 0.5, global_step)

        # update scheduler
        scheduler.step()
        lrs.append(cur_lr)

        if save_cp:
            try:
                os.mkdir(dir_checkpoint)
                logging.info('Created checkpoint directory')
            except OSError:
                pass
            
            if loss.item() < best_loss:
                torch.save(net.state_dict(), dir_checkpoint + f'CP_epoch{epoch + 1}_loss_{str(loss.item())}.pth')
                best_loss = loss.item()
                logging.info(f'Checkpoint {epoch + 1} saved ! loss (batch) = ' + str(loss.item()))
    
    # plot lr scheduler
    plt.plot(lrs , '.-', label='LambdaLR')
    plt.xlabel('epoch')
    plt.ylabel('LR')
    plt.tight_layout()
    plt.savefig('LR.png', dpi=300)

    writer.close()


def get_args():
    parser = argparse.ArgumentParser(description='Train the UNet on images and target masks',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-g', '--gpu_id', dest='gpu_id', metavar='G', type=int, default=1, help='Number of gpu')
    parser.add_argument('-u', '--unet_type', dest='unet_type', metavar='U', type=str, default='v3', help='UNet type is v1/v2/v3 (unet unet++ unet3+)')
    
    parser.add_argument('-e', '--epochs', metavar='E', type=int, default=10, help='Number of epochs', dest='epochs')
    parser.add_argument('-b', '--batch-size', metavar='B', type=int, nargs='?', default=2, help='Batch size', dest='batchsize')
    parser.add_argument('-l', '--learning-rate', metavar='LR', type=float, nargs='?', default=0.001, help='Learning rate', dest='lr')
    
    parser.add_argument('-f', '--load', dest='load', type=str, default=False, help='Load model from a .pth file')
    parser.add_argument('-s', '--scale', dest='scale', type=float, default=0.5, help='Downscaling factor of the images')
    parser.add_argument('-v', '--validation', dest='val', type=float, default=10.0, help='Percent of the data that is used as validation (0-100)')
    return parser.parse_args()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
    args = get_args()
    gpu_id = args.gpu_id 
    unet_type = args.unet_type 

    os.environ['CUDA_VISIBLE_DEVICES'] = str(gpu_id)
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    logging.info(f'Using device {device}')

    # Change here to adapt to your data
    # n_channels=3 for RGB images
    # n_classes is the number of probabilities you want to get per pixel
    #   - For 1 class and background, use n_classes=1
    #   - For 2 classes, use n_classes=1
    #   - For N > 2 classes, use n_classes=N
    if unet_type == 'v2':
        net = UNet2Plus(n_channels=3, n_classes=1)
    elif unet_type == 'v3':
        net = UNet3Plus(n_channels=3, n_classes=4)
        #net = UNet3Plus_DeepSup(n_channels=3, n_classes=1)
        #net = UNet3Plus_DeepSup_CGM(n_channels=3, n_classes=1)
    else:
        net = UNet(n_channels=3, n_classes=1)

    logging.info(f'Network:\n'
                 f'\t{net.n_channels} input channels\n'
                 f'\t{net.n_classes} output channels (classes)\n')
                 #f'\t{'Bilinear' if net.bilinear else 'Dilated conv'} upscaling')

    if args.load:
        net.load_state_dict(torch.load(args.load, map_location=device))
        logging.info(f'Model loaded from {args.load}')

    net.to(device=device)
    # faster convolutions, but more memory
    # cudnn.benchmark = True
    try:
        train_net(unet_type=unet_type, net=net, epochs=args.epochs, batch_size=args.batchsize, 
                  lr=args.lr, device=device, img_scale=args.scale, val_percent=args.val / 100)
    except KeyboardInterrupt:
        torch.save(net.state_dict(), 'INTERRUPTED.pth')
        logging.info('Saved interrupt')
        try:
            sys.exit(0)
        except SystemExit:
            os.exit(0)



import tensorflow as tf
import datetime
import numpy as np

# 예시를 위한 간단한 모델 정의
class SimpleModel(tf.keras.Model):
    def __init__(self):
        super(SimpleModel, self).__init__()
        self.dense1 = tf.keras.layers.Dense(64, activation='relu')
        self.dense2 = tf.keras.layers.Dense(10, activation='softmax')

    def call(self, x):
        x = self.dense1(x)
        x = self.dense2(x)
        return x

# 데이터 생성
x_train = np.random.randn(100, 32).astype(np.float32)
y_train = np.random.randint(0, 10, 100)

# 모델 및 옵티마이저 초기화
model = SimpleModel()
optimizer = tf.keras.optimizers.Adam()

# 손실 함수 정의
loss_object = tf.keras.losses.SparseCategoricalCrossentropy()

# TensorBoard 로그 디렉토리 설정
log_dir = "/HDD/etc/tensorboard/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
summary_writer = tf.summary.create_file_writer(log_dir)

# 훈련 스텝 함수 정의
@tf.function
def train_step(x, y, step):
    with tf.GradientTape() as tape:
        predictions = model(x, training=True)
        loss = loss_object(y, predictions)
    gradients = tape.gradient(loss, model.trainable_variables)
    optimizer.apply_gradients(zip(gradients, model.trainable_variables))
    
    with summary_writer.as_default():
        # 손실 기록
        tf.summary.scalar('loss', loss, step=step)
        
        # 가중치 및 그래디언트 기록
        for i, (weight, grad) in enumerate(zip(model.trainable_variables, gradients)):
            tf.summary.histogram(f'layer_{i+1}_weights', weight, step=step)
            tf.summary.histogram(f'layer_{i+1}_gradients', grad, step=step)
    
    return loss

def train_model(epochs, batch_size=32):
    step = 0
    dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train))
    dataset = dataset.batch(batch_size)
    
    for epoch in range(epochs):
        print(f'Starting epoch {epoch+1}')
        
        for x_batch, y_batch in dataset:
            loss = train_step(x_batch, y_batch, step)
            step += 1
        
        # 활성화 값 기록
        with summary_writer.as_default():
            activations = model(tf.convert_to_tensor(x_train))
            tf.summary.histogram('activations', activations, step=epoch)
        
    summary_writer.close()

# 모델 훈련 실행
train_model(epochs=2, batch_size=32)



# Copyright (c) ByteDance, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""
Mostly copy-paste from BEiT library:
https://github.com/microsoft/unilm/blob/master/beit/semantic_segmentation/mmcv_custom/layer_decay_optimizer_constructor.py
"""

import json

from mmcv.runner import OPTIMIZER_BUILDERS, DefaultOptimizerConstructor
from mmcv.runner import get_dist_info

def get_num_layer_for_vit(var_name, num_max_layer):
    if var_name in ("backbone.cls_token", "backbone.mask_token", "backbone.pos_embed"):
        return 0
    elif var_name.startswith("backbone.patch_embed"):
        return 0
    elif var_name.startswith("backbone.blocks"):
        layer_id = int(var_name.split('.')[2])
        return layer_id + 1
    else:
        return num_max_layer - 1


@OPTIMIZER_BUILDERS.register_module()
class LayerDecayOptimizerConstructor(DefaultOptimizerConstructor):
    def _validate_cfg(self):
        if 'custom_keys' in self.paramwise_cfg:
            if not isinstance(self.paramwise_cfg['custom_keys'], dict):
                raise TypeError(
                    'If specified, custom_keys must be a dict, '
                    f'but got {type(self.paramwise_cfg["custom_keys"])}')
            if self.base_wd is None:
                for key in self.paramwise_cfg['custom_keys']:
                    if 'decay_mult' in self.paramwise_cfg['custom_keys'][key]:
                        raise ValueError('base_wd should not be None')

        # get base lr and weight decay
        # weight_decay must be explicitly specified if mult is specified
        if ('bias_decay_mult' in self.paramwise_cfg
                or 'norm_decay_mult' in self.paramwise_cfg
                or 'dwconv_decay_mult' in self.paramwise_cfg):
            if self.base_wd is None:
                raise ValueError('base_wd should not be None')

    def add_params(self, params, module, prefix='', is_dcn_module=None):
        """Add all parameters of module to the params list.
        The parameters of the given module will be added to the list of param
        groups, with specific rules defined by paramwise_cfg.
        Args:
            params (list[dict]): A list of param groups, it will be modified
                in place.
            module (nn.Module): The module to be added.
            prefix (str): The prefix of the module
            is_dcn_module (int|float|None): If the current module is a
                submodule of DCN, `is_dcn_module` will be passed to
                control conv_offset layer's learning rate. Defaults to None.
        """
        parameter_groups = {}
        print(self.paramwise_cfg)
        if isinstance(self.paramwise_cfg, (list, tuple, set)):
            assert len(self.paramwise_cfg) == 2
            num_layers, layer_decay_rate = self.paramwise_cfg
        else:
            num_layers = self.paramwise_cfg.get('num_layers')
            layer_decay_rate = self.paramwise_cfg.get('layer_decay_rate')
        num_layers = num_layers + 2
        print("Build LayerDecayOptimizerConstructor %f - %d" % (layer_decay_rate, num_layers))
        weight_decay = self.base_wd

        for name, param in module.named_parameters():
            if not param.requires_grad:
                continue  # frozen weights
            if (
                len(param.shape) == 1 or name.endswith(".bias") or 
                name.endswith("_token") or name.endswith("pos_embed")
            ):
                group_name = "no_decay"
                this_weight_decay = 0.
            else:
                group_name = "decay"
                this_weight_decay = weight_decay

            layer_id = get_num_layer_for_vit(name, num_layers)
            group_name = "layer_%d_%s" % (layer_id, group_name)

            if group_name not in parameter_groups:
                scale = layer_decay_rate ** (num_layers - layer_id - 1)

                parameter_groups[group_name] = {
                    "weight_decay": this_weight_decay,
                    "params": [],
                    "param_names": [], 
                    "lr_scale": scale, 
                    "group_name": group_name, 
                    "lr": scale * self.base_lr, 
                }

            parameter_groups[group_name]["params"].append(param)
            parameter_groups[group_name]["param_names"].append(name)
        rank, _ = get_dist_info()
        if rank == 0:
            to_display = {}
            for key in parameter_groups:
                to_display[key] = {
                    "param_names": parameter_groups[key]["param_names"], 
                    "lr_scale": parameter_groups[key]["lr_scale"], 
                    "lr": parameter_groups[key]["lr"], 
                    "weight_decay": parameter_groups[key]["weight_decay"], 
                }
            print("Param groups = %s" % json.dumps(to_display, indent=2))
        
        # state_dict = module.state_dict()
        # for group_name in parameter_groups:
        #     group = parameter_groups[group_name]
        #     for name in group["param_names"]:
        #         group["params"].append(state_dict[name])
        params.extend(parameter_groups.values())



import os
import os.path as osp 
import glob 
import json 
import numpy as np
from tqdm import tqdm 
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns 
sns.set_style('darkgrid')

from visionsuite.utils.dataset.formats.labelme.charts import draw_dist_chart, draw_pie_chart
from pdf import create_pdf

def analyze_labelme(input_dir, output_dir, project_name, title, subtitle):
    if not osp.exists(output_dir):
        os.mkdir(output_dir)
        
    modes = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "**")) if not osp.isfile(folder)]

    if len(modes) == 0:
        modes = ['./']

    num_images = {}
    labels_by_mode = {}
    labels_by_image = {}
    objects_by_label = {}

    for mode in modes:
        json_files = glob.glob(osp.join(input_dir, mode, '*.json'))
        
        mode_labels = {}
        count = 0
        for json_file in tqdm(json_files, desc=mode):
            count += 1
            image_labels = {}
            filename = osp.split(osp.splitext(json_file)[0])[-1]
            with open(json_file, 'r') as jf:
                anns = json.load(jf)['shapes']
            
            xs, ys = [], []
            for ann in anns:
                label = ann['label']
                # labels by mode
                if label in mode_labels.keys():
                    mode_labels[label] += 1
                else:
                    mode_labels[label] = 1

                # labels by image
                if label in image_labels.keys():
                    image_labels[label] += 1
                else:
                    image_labels[label] = 1
                    
                # objects
                points = ann['points']
                for point in points:
                    xs.append(point[0])
                    ys.append(point[1])
                width = np.max(xs) - np.min(xs)
                height = np.max(ys) - np.min(ys)
                if label in objects_by_label.keys():
                    objects_by_label[label]['width'].append(width)
                    objects_by_label[label]['height'].append(height)
                else:
                    objects_by_label[label] = {'width': [width], 'height': [height]}
                    
            image_labels.update({"mode": mode})
            labels_by_image.update({filename: image_labels})
            
        num_images.update({mode: count})
        labels_by_mode.update({mode: mode_labels})
        
    # number of images
    plt.figure(figsize=(8, 6))
    bars = plt.bar(list(num_images.keys()), list(num_images.values()), color=['blue', 'green'])
    for bar, value in zip(bars, list(num_images.values())):
        plt.text(bar.get_x() + bar.get_width() / 2, bar.get_height() + 0.5, str(value), ha='center', va='bottom')
    plt.title('Number of Images')
    plt.xlabel('Dataset')
    plt.ylabel('Number of Images')
    plt.grid(True)
    plt.savefig(osp.join(output_dir, 'num_images_bar.png'))
    plt.close()
    draw_pie_chart(num_images, 'Number of Images', osp.join(output_dir, f'num_images_pie.png'))


    # labels by mode    
    for key, val in labels_by_mode.items():
        draw_pie_chart(val, key, osp.join(output_dir, f'{key}_pie_chart.png'))
        
    # labels by image
    df_labels_by_image = pd.DataFrame(labels_by_image).T
    df_labels_by_image.name = 'filename'
    df_labels_by_image.fillna(0, inplace=True)
    df_labels_by_image.to_csv(osp.join(output_dir, 'labels_by_image.csv'))
    html_labels_by_image = df_labels_by_image.to_html()
    with open(osp.join(output_dir, 'labels_by_image.html'), 'w') as f:
        f.write(html_labels_by_image)

    plt.figure(figsize=(6, 6))  
    for key, val in objects_by_label.items():
        plt.scatter(val['width'], val['height'], marker='o', s=5, label=key)
    plt.title('Object XYs')  
    plt.xlabel('Width') 
    plt.ylabel('Height')
    plt.grid(True) 
    plt.legend()
    plt.savefig(osp.join(output_dir, 'objects_xy_by_labels.png'))
    plt.close()

    txt_size = open(osp.join(output_dir, 'min_max_size.txt'), 'w')
    for key, val in objects_by_label.items():
        plt.figure(figsize=(6, 6))  
        plt.scatter(val['width'], val['height'], marker='o', s=5, label=key)
        plt.figtext(0.01, 0, f"Min. height : {np.min(val['height'])}, Max height: {np.max(val['height'])}", fontsize=10)
        plt.figtext(0.01, 0.02, f"Min. width : {np.min(val['width'])}, Max width: {np.max(val['width'])}", fontsize=10)
        plt.title(f'{key} XYs')  
        plt.xlabel('Width') 
        plt.ylabel('Height')
        plt.grid(True) 
        # plt.legend()
        plt.savefig(osp.join(output_dir, f'objects_xy_by_{key}.png'))
        plt.close()

        txt_size.write(f"{key}: \n")
        txt_size.write(f"   * min. height: {np.min(val['height'])}\n")
        txt_size.write(f"   * min. width: {np.min(val['width'])}\n\n")
    txt_size.close()        
        
        
        
    for key, val in objects_by_label.items():
        draw_dist_chart(val, key, osp.join(output_dir, f'objects_wh_{key}.png'))


    create_pdf(title, subtitle, output_dir)

if __name__ == '__main__':
    # input_dir = '/DeepLearning/etc/_athena_tests/benchmark/sungwoo/inner_body/split_dataset'
    # output_dir = '/DeepLearning/etc/_athena_tests/benchmark/sungwoo/inner_body/reports'

    # project_name = 'Sungwoo-inner-body'
    # title = "Dataset Analysis"
    # subtitle = f"The dataset is {project_name}"
    
    # analyze_labelme(input_dir, output_dir, project_name, title, subtitle)
    
    input_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/split_dataset'
    output_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/split_dataset'

    project_name = 'SFAW'
    title = "Dataset Analysis"
    subtitle = f"The dataset is {project_name}"
    
    analyze_labelme(input_dir, output_dir, project_name, title, subtitle)
    
    